%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXC:Mips-instrs.sl
% Title:          Instructions for Mips PSL
% Author:         Winfried Neun
% Created:        September 15, 1989
% Status:         Experimental
% Version for MIPS 64 derived from MIPS 32
%
% Copyright (c) 1989, 1997 Konrad-Zuse-Zentrum Berlin, All Rights Reserved
% Copyright (c) 1989, Konrad-Zuse-Zentrum Berlin, All Rights Reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
     
(compiletime (load fasl-decls common))

(global '(*writingasmfile))

(fluid '(labeloffsets!* currentoffset* codesize* codebase* entries*
	 forwardinternalreferences* lapreturnvalue*  *writingfaslfile
	 initoffset* *pgwd *pwrds label-alist*))

(compiletime (flag '(
ENTRY-GEN EXPAND-WORD EXPAND-STRING EXPAND-FLOAT DEPOSITINSTRUCTION 
FORMAT-MATCHER RD-MATCH RS-MATCH RT-MATCH FD-MATCH FS-MATCH FT-MATCH 
SHAMT-MATCH BASED-MATCH IMM-MATCH SIMM-MATCH OFFSET-MATCH TARGET-MATCH 
MATCHERROR *I-FORMAT* *J-FORMAT* *R-FORMAT* *FR-FORMAT* *FR-FORMAT-2* 
*FI-FORMAT* MAKE-I-TYPE MAKE-J-TYPE MAKE-R-TYPE MAKE-FR-TYPE MAKE-FR2-TYPE 
MAKE-FI-TYPE NOP-GEN MOVEX-GEN MOVEX-LOAD MOVEX-STORE RELOCATABLE-MOVEX-GEN 
RELOCATABLE-MOVEX-STORE RELOCATABLE-MOVEX-LOAD) 'iinternalfunction))

%%|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%k
%%      Assembler code generator
%%
%%|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

(compiletime 
(dm define-instruction (u)
 (prog (name length result)
  (setq u (cdr u)) 
  (setq name (first u))
  (setq length (second u))
  (push (list 'put (mkquote name) (mkquote '*op-length*) length) result)
  (setq u (cddr u))    
  (setq result
     (append (mapcar u (function 
	 (lambda (x) (list 'put (mkquote name)
		                (mkquote (car x)) (mkquote (cdr x))))))
	result))
  (return(cons 'progn result))))
)

%%      Pseudo-Instructions

(define-instruction  *ENTRY 0 (*apply* entry-gen))
%
%       (*ENTRY name type argn)
%
(de entry-gen (instruction name type argn)
  (saveentry (list instruction name type argn)))

%
%       (WORD value ......)
%
(define-instruction WORD      8 (*apply* expand-word))
(define-instruction FullWORD  8 (*apply* expand-word))
(define-instruction STRING    0 (*apply* expand-string))
		            %will be treated by onelaplength
(define-instruction Float     8 (*apply* expand-float))

(de expand-word (in x)        (store-word-expression x))

(de expand-string (in x)
	(for (from i 0 (size x))
	     (do (setf (byte (wplus2 codebase* currentoffset*) i) (indx x i))))
	(setf (byte (wplus2 codebase* currentoffset*) (wplus2 (size x) 1)) 0)
	(setq currentoffset* (wplus2 (wtimes2 8 % addressingunitsperitem
				      (strpack (size x))) currentoffset*)))

(de expand-float (in x)         (store-double x ))

%%    The Instruction processor

(de depositinstruction (in)
   (prog (name format spec oldcurrent lng oldbto)
 
      (setq oldcurrent currentoffset*)
      (when *writingfaslfile (setq oldbto bittableoffset*))
      (when (atom in) (return nil)) % skip labels
      (setq name (first in))
      (setq lng (cond ((eq name 'string) % hier
		        (wtimes2 8 %addressingunitsperitem
				 (strpack (size (second in)))) )
                      ((eq name 'oddalign) 0)
                      ((eq name 'wordalign)
                        (if (eq currentoffset* (wand currentoffset* (wnot 7)))
                                0 4))
		      (t (get name '*op-length*))))
      (when *writingfaslfile (updatebittable (wshift lng 2) 0) % clear bittable 
		             (setq bittableoffset* oldbto))
      
      (cond ((get name '*apply*) (apply (car (get name '*apply*)) in))
		          % escape for special pseudo instructions
	    ((setq spec (get name '*special*))
		 (format-matcher in (get name '*r-format*)
		         (list (cons 'funct (car spec))) '*r-format*))
	    ((setq spec (get name '*bcond*))
		(setq format (get name '*i-format*))
		(format-matcher in format (list (cons 'rt (car spec)))
		                '*i-format*)) 
	    ((setq format (get name '*i-format*))
		 (format-matcher in format nil '*i-format*))
	    ((setq format (get name '*j-format*))
		         (format-matcher in format nil '*j-format*))
	    ((setq format (get name '*fr-format*))  %float instructions
		  (format-matcher in (cdr format)
		         (list (cons 'fmt (car format))
		               (cons 'funct (car (get name '*cop1*))))
		         '*fr-format*))
	    ((setq format (get name '*fr-format-2*))  %float instructions
		  (format-matcher in (cdr format)
		         (list (cons 'fmt (car format))
		               (cons 'funct (car (or (get name '*cop0*)
						     (get name '*cop1*)))))
		         '*fr-format-2*))
	    ((setq format (get name '*fi-format*))  %float instructions
		  (format-matcher in (cdr format)
		         (list (cons 'fmt (car format)))
		         '*fi-format*))
   (t (error 4711 (list "Wrong Mips instruction :" in))))
  % 
  % update currentoffset*
  %
    (setq currentoffset* (plus2 oldcurrent lng)) 
    (when *writingfaslfile (setq bittableoffset* (plus2 oldbto lng )))
) )

(de format-matcher (ins format initlist formfkt) 
   (prog (in name item fmtitem alist appl)
       (setq alist 
	     (append initlist
		`((rs . 0) (rt . 0) (rd . 0) (shamt . 0) (imm . 0)
		  (fs . 0) (ft . 0) (fd . 0) (fmt . 0) % float intrs
		 (funct . 0) (offset . 0) (based . 0) (simm  . 0))))
       (setq name (car ins))
       (setq in (cdr ins)) 
	(while (pairp in)
	 (setq item (car in))
	 (setq fmtitem (car format))
	 (cond  ((numberp fmtitem)  nil)
      		((eq (tag fmtitem) 255) nil) %temporary
		((and (idp fmtitem)
		      (setq appl (get fmtitem 'matchfn)))
		      (setq alist (apply appl (list (car in) alist))))
		(t (error 4711 (list "wrong instruction " ins))))
	 (setq in (cdr in))
	 (setq format (cdr format))
	 (when (and (not (pairp format)) in)
		(error 4711 (list "Mips wrong instruction :" ins))))

	(while format
	   (when (idp (car format))
	       (error 4711 (list "Mips instruction too short :" ins)))
	   (setq format (cdr format)))

   (apply formfkt (list name alist))
)  )
   
(put 'rd 'matchfn 'rd-match)

(de rd-match (r alist)
    (if (regp r) (cons (cons 'rd r) alist)
     (matcherror "Register" r)))

(put 'rs 'matchfn 'rs-match)

(de rs-match (x alist)
    (if (regp x) (cons (cons 'rs x) alist)
     (matcherror "Register" x)))

(put 'rt 'matchfn 'rt-match)

(de rt-match (x alist)
    (if (regp x) (cons (cons 'rt x) alist)
     (matcherror "Register" x)))

(put 'fd 'matchfn 'fd-match)

(de fd-match (r alist)
    (if (fregp r) (cons (cons 'fd r) alist)
     (matcherror "Float Register" r)))

(put 'fs 'matchfn 'fs-match)

(de fs-match (x alist)
    (if (fregp x) (cons (cons 'fs x) alist)
     (matcherror "Float Register" x)))

(put 'ft 'matchfn 'ft-match)

(de ft-match (x alist)
    (if (fregp x) (cons (cons 'ft x) alist)
     (matcherror "Float Register" x)))

(put 'shamt 'matchfn 'shamt-match)

(de shamt-match (x alist)
    (if (and (posintp x) (wlessp x 65))  (cons (cons 'shamt x) alist)
     (matcherror "Shift amount"x)))

(put 'based 'matchfn 'based-match)

(de based-match (x alist)
   (if (atom x)   (cons (cons 'immediate (addressvalue x))
		         (cons (cons 'rs '(reg null)) alist))
      (cond ((eq (car x) 'deferred)
		 (cons (cons 'immediate 0) (cons (cons 'rs (cadr x)) alist)))
	    ((or (eq (car x) 'indexed) (eq (car x) 'displacement))
		     (cons (cons 'immediate (addressvalue (caddr x)))
		           (cons (cons 'rs (cadr x)) alist)))
	    ((eq (car x) 'entryaddress)
		(cons (cons 'immediate 
		        (wand (addressvalue (list 'xidloc (cadr x))) 16#ffff))
		        (cons (cons 'rs '(reg symfnccenter)) alist)))
	    ((eq (car x) 'valueaddress)
		(cons (cons 'immediate 
		        (wand (addressvalue (list 'xidloc (cadr(cadr x))))
		                 16#ffff))
		        (cons (cons 'rs '(reg symvalcenter)) alist)))
	    ((eq (car x) 'loloc)
                (cons (cons 'immediate
                        (wand (addressvalue x) 16#ffff))
                        (cons (cons 'rs '(reg symvalcenter)) alist)))
	    (t (matcherror "based" x)))))

(put 'imm 'matchfn 'imm-match)

(de imm-match (x alist)
  (prog (add)
     (return  
	(if (setq add (addressvalue x))  
		(cons (cons 'immediate (wand add 16#ffff)) alist)
	     (matcherror "immediate" x)))))

(put 'simm 'matchfn 'simm-match)

(de simm-match (x alist)
  (prog (add)
     (return  
	(if (setq add (addressvalue x))  
		        (cons (cons 'immediate (wand add 16#ffff)) alist)
     (matcherror "signed immediate" x))))) 

(put 'offset 'matchfn 'offset-match)

(de offset-match (x alist) % branch address
  (prog (add)
     (return  
      (if (setq add (addressvalue x))  
	(cons (cons 'immediate
		(wand (wplus2 -1 
			(wshift (wdifference add currentoffset*) -2))
		    16#ffff)) 
		alist) % word addresses
     (matcherror "offset" x)))))

(put 'target 'matchfn 'target-match)

(de target-match (x alist) % jump address
  (prog (add)
     (return  
	(if (setq add (jumpaddressvalue x)) 
		(cons (cons 'target (wshift (wshift add 6) -8)) alist) 
                          % throw out hightest bits (see J instruction!
     (matcherror "target" x))))) 

(de matcherror (expected got)
    (error 4711 (list "Mismatch of Arguments (" expected "expected) :" got))) 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% The formats : r, i, j and fr for floating r format 

%%      i-type instructions

(compiletime (ds valueofreg (x)
		(if (pairp (setq cdrass (cdr (atsoc x alist))))
		     (get-register-code cdrass) cdrass)))
  
(de *i-format* (name alist)
   (prog (cdrass)
     (return
	(make-i-type (or (get name '*opcode*)
		         (and (get name '*bcond*) (ncons 1)))
		     (valueofreg 'rs)
		     (if (get name '*cop1*) (valueofreg 'ft) %lwc1 and swc1
		                (valueofreg 'rt))
		     (if (pairp (setq cdrass (cdr (atsoc 'immediate alist))))
		        (car cdrass) cdrass)
		     )) ))
 
%%      j-type instructions
 
(de *j-format* (name alist) 
	(make-j-type (get name '*opcode*) 
		     (cdr (atsoc 'target alist))))
 
%%      r-type instructions
 
(de *r-format* (name alist)
   (prog (cdrass)
     (return
	(make-r-type (or (get name '*opcode*) (ncons 0)) % special opcode
		     (valueofreg 'rs)
		     (valueofreg 'rt)
		     (valueofreg 'rd)
		     (if (pairp (setq cdrass (cdr (atsoc 'shamt alist))))  
		        (car cdrass) cdrass)  
		     (if (pairp (setq cdrass (cdr (atsoc 'funct alist))))  
		        (car cdrass) cdrass)
		     ))
 ))

(de *fr-format* (name alist) 
   (prog (cdrass)
     (return
       (make-fr-type (or (get name '*opcode*) (ncons 8#21)) % COP1 opcode
		     (if (pairp (setq cdrass (cdr (atsoc 'fmt alist))))
		        (car cdrass) cdrass)
		     (valueofreg 'ft)
		     (valueofreg 'fs)
		     (valueofreg 'fd)
		     (if (pairp (setq cdrass (cdr (atsoc 'funct alist))))  
		        (car cdrass) cdrass)
		     ))
 ))

(de *fr-format-2* (name alist) 
   (prog (cdrass)
     (return
       (make-fr2-type (or (get name '*opcode*) 
                          (and (get name '*cop0*) (ncons 8#20))
                          (ncons 8#21)) % COP1 opcode
		     (if (pairp (setq cdrass (cdr (atsoc 'fmt alist))))
		        (car cdrass) cdrass)
		     (valueofreg 'rt)
		     (valueofreg 'fs)
		     (if (pairp (setq cdrass (cdr (atsoc 'funct alist))))  
		        (car cdrass) cdrass)
		     ))
 ))

(de *fi-format* (name alist) %(name rs rt immediate)
   (prog (cdrass)
     (return
	(make-fi-type (ncons 8#21)
		      (if (pairp (setq cdrass (cdr (atsoc 'fmt alist))))
		        (car cdrass) cdrass)
		      (if (pairp (setq cdrass (cdr (atsoc 'immediate alist))))
		        (car cdrass) cdrass)))
))

% The make functions
 
(de make-i-type (op rs rt imm)
     (let ((code (wor (wshift (car op) 26)
		      (wor (wshift rs 21)
		           (wor (wshift rt 16)
		                (wand imm 16#ffff))))))
       (store-word code)))
 
(de make-j-type (op target)
     (store-word (wor (wshift (car op) 26) target )))
 
(de make-r-type (op rs rt rd shamt funct)
     (let ((code (wor (wshift (car op) 26)
		      (wor (wshift rs 21)
		           (wor (wshift rt 16)
		                (wor (wshift rd 11)
		                     (wor (wshift shamt 6)
		                          funct)))))))
      (store-word code)))

(de make-fr-type (op fmt ft fs fd funct)
     (let ((code (wor (wshift (car op) 26)
		      (wor (wshift (wplus2 fmt 16) 21)
		           (wor (wshift ft 16)
		                (wor (wshift fs 11)
		                     (wor (wshift fd 6)
		                          funct)))))))
      (store-word code)))

(de make-fr2-type (op fmt rt fs)
     (let ((code (wor (wshift (car op) 26)
		      (wor (wshift fmt 21)
		           (wor (wshift rt 16)
	                        (wshift fs 11))))))
      (store-word code)))

(de make-fi-type (op fmt imm)
     (let ((code (wor (wshift (car op) 26)
		         (wor (wshift fmt 16) (wand imm 16#ffff)))))
       (store-word code)))
 
%%    The Instruction definitions

(define-instruction sll     4 (*special* 8#00) (*r-format* rd rt shamt)) 
(define-instruction srl     4 (*special* 8#02) (*r-format* rd rt shamt)) 
(define-instruction sra     4 (*special* 8#03) (*r-format* rd rt shamt)) 
(define-instruction sllv    4 (*special* 8#04) (*r-format* rd rt rs 0)) 
(define-instruction srlv    4 (*special* 8#06) (*r-format* rd rt rs 0)) 
(define-instruction srav    4 (*special* 8#07) (*r-format* rd rt rs 0)) 

% MIPS III

(define-instruction dsll    4 (*special* 8#70) (*r-format* rd rt shamt)) 
(define-instruction dsrl    4 (*special* 8#72) (*r-format* rd rt shamt)) 
(define-instruction dsra    4 (*special* 8#73) (*r-format* rd rt shamt)) 
(define-instruction dsllv   4 (*special* 8#24) (*r-format* rd rt rs 0)) 
(define-instruction dsrlv   4 (*special* 8#26) (*r-format* rd rt rs 0)) 
(define-instruction dsrav   4 (*special* 8#27) (*r-format* rd rt rs 0))
(define-instruction dsll32  4 (*special* 8#74) (*r-format* rd rt shamt)) 
(define-instruction dsra32  4 (*special* 8#77) (*r-format* rd rt shamt)) 
(define-instruction dsrl32  4 (*special* 8#76) (*r-format* rd rt shamt))  


(define-instruction jr      4 (*special* 8#10) (*r-format* rs ))
(define-instruction jalr    4 (*special* 8#11) (*r-format* rd rs )) 
(define-instruction syscall 4 (*special* 8#14) (*r-format* 0 0 0 0)) 
(define-instruction break   4 (*special* 8#15) (*r-format* 0 0 0 0)) 
(define-instruction mfhi    4 (*special* 8#20) (*r-format* rd )) 
(define-instruction mthi    4 (*special* 8#21) (*r-format* rd )) 
(define-instruction mflo    4 (*special* 8#22) (*r-format* rd )) 
(define-instruction mtlo    4 (*special* 8#23) (*r-format* rd ))
(define-instruction mult    4 (*special* 8#30) (*r-format* rs rt )) 
(define-instruction multu   4 (*special* 8#31) (*r-format* rs rt )) 
(define-instruction div     4 (*special* 8#32) (*r-format* rs rt )) 
(define-instruction divu    4 (*special* 8#33) (*r-format* rs rt )) 
(define-instruction add     4 (*special* 8#40) (*r-format* rd rs rt )) 
(define-instruction addu    4 (*special* 8#41) (*r-format* rd rs rt )) 
(define-instruction sub     4 (*special* 8#42) (*r-format* rd rs rt )) 
(define-instruction subu    4 (*special* 8#43) (*r-format* rd rs rt )) 
(define-instruction and     4 (*special* 8#44) (*r-format* rd rs rt )) 
(define-instruction or      4 (*special* 8#45) (*r-format* rd rs rt )) 
(define-instruction xor     4 (*special* 8#46) (*r-format* rd rs rt )) 
(define-instruction nor     4 (*special* 8#47) (*r-format* rd rs rt )) 
(define-instruction slt     4 (*special* 8#52) (*r-format* rd rs rt )) 
(define-instruction sltu    4 (*special* 8#53) (*r-format* rd rs rt )) 


% MIPS III

(define-instruction dmult   4 (*special* 8#34) (*r-format* rs rt )) 
(define-instruction dmultu  4 (*special* 8#35) (*r-format* rs rt )) 
(define-instruction ddiv    4 (*special* 8#36) (*r-format* rs rt )) 
(define-instruction ddivu   4 (*special* 8#37) (*r-format* rs rt )) 
(define-instruction dadd    4 (*special* 8#54) (*r-format* rd rs rt )) 
(define-instruction daddu   4 (*special* 8#55) (*r-format* rd rs rt )) 
(define-instruction dsub    4 (*special* 8#56) (*r-format* rd rs rt )) 
(define-instruction dsubu   4 (*special* 8#57) (*r-format* rd rs rt )) 

(define-instruction bltz    4 (*bcond* 8#00) (*i-format* rs offset)) 
(define-instruction bgez    4 (*bcond* 8#01) (*i-format* rs offset)) 
(define-instruction bltzal  4 (*bcond* 8#20) (*i-format* rs offset)) 
(define-instruction bgezal  4 (*bcond* 8#21) (*i-format* rs offset)) 

(define-instruction special 4 (*opcode* 8#00) (*expand* *special*))
(define-instruction bcond   4 (*opcode* 8#01) (*expand* *bcond*)) 
(define-instruction j       4 (*opcode* 8#02) (*j-format* target)) 
(define-instruction jal     4 (*opcode* 8#03) (*j-format* target)) 
(define-instruction beq     4 (*opcode* 8#04) (*i-format* rt rs offset)) 
(define-instruction bne     4 (*opcode* 8#05) (*i-format* rt rs offset)) 
(define-instruction blez    4 (*opcode* 8#06) (*i-format* rs offset)) 
(define-instruction bgtz    4 (*opcode* 8#07) (*i-format* rs offset)) 
(define-instruction addi    4 (*opcode* 8#10) (*i-format* rt rs simm)) 
(define-instruction addiu   4 (*opcode* 8#11) (*i-format* rt rs simm)) 

% MIPS III

(define-instruction daddi   4 (*opcode* 8#30) (*i-format* rt rs simm))
(define-instruction daddiu  4 (*opcode* 8#31) (*i-format* rt rs simm)) 

(define-instruction slti    4 (*opcode* 8#12) (*i-format* rt rs simm))
(define-instruction sltiu   4 (*opcode* 8#13) (*i-format* rt rs simm)) 
(define-instruction andi    4 (*opcode* 8#14) (*i-format* rt rs imm)) 
(define-instruction ori     4 (*opcode* 8#15) (*i-format* rt rs imm)) 
(define-instruction xori    4 (*opcode* 8#16) (*i-format* rt rs imm)) 
(define-instruction lui     4 (*opcode* 8#17) (*i-format* rt imm)) 
(define-instruction cop0    4 (*opcode* 8#20) (*expand* *cop0*)) 
(define-instruction cop1    4 (*opcode* 8#21) (*expand* *cop1*)) 
(define-instruction cop2    4 (*opcode* 8#22) (*expand* *cop2*)) 
(define-instruction cop3    4 (*opcode* 8#23) (*expand* *cop3*)) 
(define-instruction lb      4 (*opcode* 8#40) (*i-format* rt based)) 
(define-instruction lh      4 (*opcode* 8#41) (*i-format* rt based)) 
(define-instruction lwl     4 (*opcode* 8#42) (*i-format* rt based)) 
(define-instruction lw      4 (*opcode* 8#43) (*i-format* rt based)) 
(define-instruction lwr     4 (*opcode* 8#46) (*i-format* rt based)) 
(define-instruction lwu     4 (*opcode* 8#47) (*i-format* rt based)) 

% MIPS III

(define-instruction ld      4 (*opcode* 8#67) (*i-format* rt based))  
(define-instruction ldl     4 (*opcode* 8#32) (*i-format* rt based)) 
(define-instruction ldr     4 (*opcode* 8#33) (*i-format* rt based)) 

(define-instruction sd      4 (*opcode* 8#77) (*i-format* rt based))  
(define-instruction sdl     4 (*opcode* 8#54) (*i-format* rt based)) 
(define-instruction sdr     4 (*opcode* 8#55) (*i-format* rt based)) 

(define-instruction lbu     4 (*opcode* 8#44) (*i-format* rt based))
(define-instruction lhu     4 (*opcode* 8#45) (*i-format* rt based)) 
(define-instruction lwr     4 (*opcode* 8#46) (*i-format* rt based)) 
(define-instruction sb      4 (*opcode* 8#50) (*i-format* rt based)) 
(define-instruction sh      4 (*opcode* 8#51) (*i-format* rt based))
(define-instruction swl     4 (*opcode* 8#52) (*i-format* rt based))
(define-instruction swr     4 (*opcode* 8#56) (*i-format* rt based))
(define-instruction sw      4 (*opcode* 8#53) (*i-format* rt based)) 
(define-instruction lwc0    4 (*opcode* 8#60) (*i-format* rt based)) 
(define-instruction lwc1  4 (*opcode* 8#61) (*i-format* ft based) (*cop1* t))
(define-instruction lwc2    4 (*opcode* 8#62) (*i-format* rt based)) 
(define-instruction lwc3    4 (*opcode* 8#63) (*i-format* rt based)) 
(define-instruction swc0    4 (*opcode* 8#70) (*i-format* rt based)) 
(define-instruction swc1  4 (*opcode* 8#71) (*i-format* ft based) (*cop1* t))
(define-instruction swc2    4 (*opcode* 8#72) (*i-format* rt based)) 
(define-instruction swc3    4 (*opcode* 8#73) (*i-format* rt based)) 


(define-instruction ldc0    4 (*opcode* 8#64) (*i-format* rt based)) 
(define-instruction ldc1  4 (*opcode* 8#65) (*i-format* ft based) (*cop1* t))
(define-instruction ldc2    4 (*opcode* 8#66) (*i-format* rt based)) 
(define-instruction ldc3    4 (*opcode* 8#67) (*i-format* rt based)) 
(define-instruction sdc0    4 (*opcode* 8#74) (*i-format* rt based)) 
(define-instruction sdc1  4 (*opcode* 8#75) (*i-format* ft based) (*cop1* t))
(define-instruction sdc2    4 (*opcode* 8#76) (*i-format* rt based)) 
(define-instruction sdc3    4 (*opcode* 8#77) (*i-format* rt based)) 

% The floating point (FPA) instructions via COP1

(define-instruction add!.s    4 (*cop1* 8#00) (*fr-format* 0 fd fs ft)) 
(define-instruction add!.d    4 (*cop1* 8#00) (*fr-format* 1 fd fs ft)) 
(define-instruction add!.w    4 (*cop1* 8#00) (*fr-format* 1 fd fs ft)) 
(define-instruction sub!.s    4 (*cop1* 8#01) (*fr-format* 0 fd fs ft)) 
(define-instruction sub!.d    4 (*cop1* 8#01) (*fr-format* 1 fd fs ft)) 
(define-instruction sub!.w    4 (*cop1* 8#01) (*fr-format* 1 fd fs ft)) 
(define-instruction mul!.s    4 (*cop1* 8#02) (*fr-format* 0 fd fs ft)) 
(define-instruction mul!.d    4 (*cop1* 8#02) (*fr-format* 1 fd fs ft)) 
(define-instruction mul!.w    4 (*cop1* 8#02) (*fr-format* 1 fd fs ft)) 
(define-instruction div!.s    4 (*cop1* 8#03) (*fr-format* 0 fd fs ft)) 
(define-instruction div!.d    4 (*cop1* 8#03) (*fr-format* 1 fd fs ft)) 
(define-instruction div!.w    4 (*cop1* 8#03) (*fr-format* 1 fd fs ft)) 
(define-instruction abs!.s    4 (*cop1* 8#05) (*fr-format* 0 fd fs )) 
(define-instruction abs!.d    4 (*cop1* 8#05) (*fr-format* 1 fd fs )) 
(define-instruction abs!.w    4 (*cop1* 8#05) (*fr-format* 1 fd fs )) 
(define-instruction mov!.s    4 (*cop1* 8#06) (*fr-format* 0 fd fs ft)) 
(define-instruction mov!.d    4 (*cop1* 8#06) (*fr-format* 1 fd fs ft)) 
(define-instruction mov!.w    4 (*cop1* 8#06) (*fr-format* 1 fd fs ft)) 
(define-instruction neg!.s    4 (*cop1* 8#07) (*fr-format* 0 fd fs ft)) 
(define-instruction neg!.d    4 (*cop1* 8#07) (*fr-format* 1 fd fs ft)) 
(define-instruction neg!.w    4 (*cop1* 8#07) (*fr-format* 1 fd fs ft)) 
(define-instruction cvt!.s!.s  4 (*cop1* 8#40) (*fr-format* 0 fd fs)) 
(define-instruction cvt!.s!.d  4 (*cop1* 8#40) (*fr-format* 1 fd fs)) 
(define-instruction cvt!.s!.w  4 (*cop1* 8#40) (*fr-format* 4 fd fs)) 
(define-instruction cvt!.s!.l  4 (*cop1* 8#40) (*fr-format* 5 fd fs)) 
(define-instruction cvt!.d!.s  4 (*cop1* 8#41) (*fr-format* 0 fd fs)) 
(define-instruction cvt!.d!.d  4 (*cop1* 8#41) (*fr-format* 1 fd fs)) 
(define-instruction cvt!.d!.w  4 (*cop1* 8#41) (*fr-format* 4 fd fs)) 
(define-instruction cvt!.d!.l  4 (*cop1* 8#41) (*fr-format* 5 fd fs)) 
(define-instruction cvt!.w!.s  4 (*cop1* 8#44) (*fr-format* 0 fd fs))
(define-instruction cvt!.w!.d  4 (*cop1* 8#44) (*fr-format* 1 fd fs))
(define-instruction cvt!.w!.w  4 (*cop1* 8#44) (*fr-format* 4 fd fs))
(define-instruction cvt!.w!.l  4 (*cop1* 8#44) (*fr-format* 5 fd fs))
(define-instruction c!.f!.s    4 (*cop1* 8#60) (*fr-format* 0 fs ft)) 
(define-instruction c!.f!.d    4 (*cop1* 8#60) (*fr-format* 1 fs ft)) 
(define-instruction c!.f!.w    4 (*cop1* 8#60) (*fr-format* 4 fs ft)) 
(define-instruction c!.f!.l    4 (*cop1* 8#60) (*fr-format* 5 fs ft)) 
(define-instruction c!.un!.s   4 (*cop1* 8#61) (*fr-format* 0 fs ft)) 
(define-instruction c!.un!.d   4 (*cop1* 8#61) (*fr-format* 1 fs ft)) 
(define-instruction c!.un!.w   4 (*cop1* 8#61) (*fr-format* 4 fs ft)) 
(define-instruction c!.un!.l   4 (*cop1* 8#61) (*fr-format* 5 fs ft)) 
(define-instruction c!.eq!.s   4 (*cop1* 8#62) (*fr-format* 0 fs ft)) 
(define-instruction c!.eq!.d   4 (*cop1* 8#62) (*fr-format* 1 fs ft)) 
(define-instruction c!.eq!.w   4 (*cop1* 8#62) (*fr-format* 4 fs ft)) 
(define-instruction c!.eq!.l   4 (*cop1* 8#62) (*fr-format* 5 fs ft)) 
(define-instruction c!.ueq!.s  4 (*cop1* 8#63) (*fr-format* 0 fs ft)) 
(define-instruction c!.ueq!.d  4 (*cop1* 8#63) (*fr-format* 1 fs ft)) 
(define-instruction c!.ueq!.w  4 (*cop1* 8#63) (*fr-format* 4 fs ft)) 
(define-instruction c!.ueq!.l  4 (*cop1* 8#63) (*fr-format* 5 fs ft)) 
(define-instruction c!.olt!.s  4 (*cop1* 8#64) (*fr-format* 0 fs ft)) 
(define-instruction c!.olt!.d  4 (*cop1* 8#64) (*fr-format* 1 fs ft)) 
(define-instruction c!.olt!.w  4 (*cop1* 8#64) (*fr-format* 4 fs ft)) 
(define-instruction c!.olt!.l  4 (*cop1* 8#64) (*fr-format* 5 fs ft)) 
(define-instruction c!.ult!.s  4 (*cop1* 8#65) (*fr-format* 0 fs ft)) 
(define-instruction c!.ult!.d  4 (*cop1* 8#65) (*fr-format* 1 fs ft)) 
(define-instruction c!.ult!.w  4 (*cop1* 8#65) (*fr-format* 4 fs ft)) 
(define-instruction c!.ult!.l  4 (*cop1* 8#65) (*fr-format* 5 fs ft)) 
(define-instruction c!.ole!.s  4 (*cop1* 8#66) (*fr-format* 0 fs ft)) 
(define-instruction c!.ole!.d  4 (*cop1* 8#66) (*fr-format* 1 fs ft)) 
(define-instruction c!.ole!.w  4 (*cop1* 8#66) (*fr-format* 4 fs ft)) 
(define-instruction c!.ole!.l  4 (*cop1* 8#66) (*fr-format* 5 fs ft)) 
(define-instruction c!.ule!.s  4 (*cop1* 8#67) (*fr-format* 0 fs ft)) 
(define-instruction c!.ule!.d  4 (*cop1* 8#67) (*fr-format* 1 fs ft)) 
(define-instruction c!.ule!.w  4 (*cop1* 8#67) (*fr-format* 4 fs ft)) 
(define-instruction c!.ule!.l  4 (*cop1* 8#67) (*fr-format* 5 fs ft)) 
(define-instruction c!.sf!.s   4 (*cop1* 8#70) (*fr-format* 0 fs ft)) 
(define-instruction c!.sf!.d   4 (*cop1* 8#70) (*fr-format* 1 fs ft)) 
(define-instruction c!.sf!.w   4 (*cop1* 8#70) (*fr-format* 4 fs ft)) 
(define-instruction c!.sf!.l   4 (*cop1* 8#70) (*fr-format* 5 fs ft)) 
(define-instruction c!.ngle!.s 4 (*cop1* 8#71) (*fr-format* 0 fs ft)) 
(define-instruction c!.ngle!.d 4 (*cop1* 8#71) (*fr-format* 1 fs ft)) 
(define-instruction c!.ngle!.w 4 (*cop1* 8#71) (*fr-format* 4 fs ft)) 
(define-instruction c!.ngle!.l 4 (*cop1* 8#71) (*fr-format* 5 fs ft)) 
(define-instruction c!.seq!.s  4 (*cop1* 8#72) (*fr-format* 0 fs ft)) 
(define-instruction c!.seq!.d  4 (*cop1* 8#72) (*fr-format* 1 fs ft)) 
(define-instruction c!.seq!.w  4 (*cop1* 8#72) (*fr-format* 4 fs ft)) 
(define-instruction c!.seq!.l  4 (*cop1* 8#72) (*fr-format* 5 fs ft)) 
(define-instruction c!.ngl!.s  4 (*cop1* 8#73) (*fr-format* 0 fs ft)) 
(define-instruction c!.ngl!.d  4 (*cop1* 8#73) (*fr-format* 1 fs ft)) 
(define-instruction c!.ngl!.w  4 (*cop1* 8#73) (*fr-format* 4 fs ft)) 
(define-instruction c!.ngl!.l  4 (*cop1* 8#73) (*fr-format* 5 fs ft)) 
(define-instruction c!.lt!.s   4 (*cop1* 8#74) (*fr-format* 0 fs ft)) 
(define-instruction c!.lt!.d   4 (*cop1* 8#74) (*fr-format* 1 fs ft)) 
(define-instruction c!.lt!.w   4 (*cop1* 8#74) (*fr-format* 4 fs ft)) 
(define-instruction c!.lt!.l   4 (*cop1* 8#74) (*fr-format* 5 fs ft)) 
(define-instruction c!.nge!.s  4 (*cop1* 8#75) (*fr-format* 0 fs ft)) 
(define-instruction c!.nge!.d  4 (*cop1* 8#75) (*fr-format* 1 fs ft)) 
(define-instruction c!.nge!.w  4 (*cop1* 8#75) (*fr-format* 4 fs ft)) 
(define-instruction c!.nge!.l  4 (*cop1* 8#75) (*fr-format* 5 fs ft)) 
(define-instruction c!.le!.s   4 (*cop1* 8#76) (*fr-format* 0 fs ft)) 
(define-instruction c!.le!.d   4 (*cop1* 8#76) (*fr-format* 1 fs ft)) 
(define-instruction c!.le!.w   4 (*cop1* 8#76) (*fr-format* 4 fs ft)) 
(define-instruction c!.le!.l   4 (*cop1* 8#76) (*fr-format* 5 fs ft)) 
(define-instruction c!.ngt!.s  4 (*cop1* 8#77) (*fr-format* 0 fs ft)) 
(define-instruction c!.ngt!.d  4 (*cop1* 8#77) (*fr-format* 1 fs ft)) 
(define-instruction c!.ngt!.w  4 (*cop1* 8#77) (*fr-format* 4 fs ft)) 
(define-instruction c!.ngt!.l  4 (*cop1* 8#77) (*fr-format* 5 fs ft)) 

(define-instruction mtc1  4 (*cop1* 8#00) (*fr-format-2* 2#00100 rt fs)) 
(define-instruction mfc1  4 (*cop1* 8#00) (*fr-format-2* 2#00000 rt fs)) 
(define-instruction cfc1  4 (*cop1* 8#00) (*fr-format-2* 2#00010 rt fs)) 
(define-instruction ctc1  4 (*cop1* 8#00) (*fr-format-2* 2#00110 rt fs)) 
(define-instruction cfc0  4 (*cop0* 8#00) (*fr-format-2* 2#00010 rt fs))
(define-instruction ctc0  4 (*cop0* 8#00) (*fr-format-2* 2#00110 rt fs))
(define-instruction bc1t  4 (*cop1* 8#00) (*fi-format* 2#0100000001 offset)) 
(define-instruction bc1f  4 (*cop1* 8#00) (*fi-format* 2#0100000000 offset)) 

(define-instruction nop 4  (*apply* nop-gen))

(de nop-gen ()    (depositinstruction '(sll (reg null) (reg null) 0)))

(define-instruction oddalign  0 (*apply* expand-oddalign))

(de expand-oddalign (in) NIL)

(define-instruction wordalign  0 (*apply* expand-wordalign))

(de expand-wordalign (in)
     (when (neq (wand currentoffset* (wnot 7)) currentoffset*)
        (setq currentoffset* (wplus2 currentoffset* 4))))

%%      move-relocatable-object pseudo instructions

(define-instruction move-relocatable-object 8 (*apply* movex-gen))

(de movex-gen (name src reg) % for 32 bit items (code addresses!)
	  (cond (*writingfaslfile
              (depositinstruction `(lui ,reg (hiloc ,src)))
              (depositinstruction `(ori ,reg ,reg (loloc ,src) )))
	    % (store-word-expression src)
				%% when writing to faslfile, generate
	    % (relocatable-movex-gen reg)) %% special code for relocation.
	   (t (depositinstruction `(lui ,reg (hi ,src))) 
	      (depositinstruction `(ori ,reg ,reg (lo ,src) )))))

(define-instruction  move-relocatable-object-and-load 12 (*apply* movex-load))

(de movex-load (name reg src )
   (cond (*writingfaslfile
	   (store-word-expression  src)  %% when writing to faslfile, generate
	   (relocatable-movex-load  reg))   %% special code for relocation.
	 (t (depositinstruction `(lui ,reg (adjhi ,src)))
	    (depositinstruction `(ld ,reg (indexed ,reg (adjlo ,src)))))))

(define-instruction move-relocatable-object-and-store 12 (*apply* movex-store))

(de movex-store (name reg src cont)
  (cond (*writingfaslfile
	   (store-word-expression src) %% when writing to faslfile, generate
	   (relocatable-movex-store reg cont)) %% code for relocation.
	(t (depositinstruction `(lui ,reg (adjhi ,src)))
           (depositinstruction `(dsll32 ,reg ,reg 0))
		                          %% otherwise normal code.
	   (depositinstruction `(sd ,cont (indexed ,reg (ajdlo ,src)))))))

(de relocatable-movex-gen (reg)
  (let ((regcode (get-register-code reg)) (mov-instr 8#15))
       (store-relocatable-movex            % ori instr 
	    (wplus2 (wshift mov-instr 26)
		    (wplus2  (wshift regcode 21) (wshift regcode 16))))))

(de relocatable-movex-store (reg cont)
  (let ((regcode (get-register-code reg))
	(regcode1 (get-register-code cont))
	(mov-instr 8#77))  %sd
       (store-relocatable-movex 
	    (wplus2 (wshift mov-instr 26)
		    (wplus2  (wshift regcode 21) (wshift regcode1 16))))))

(de relocatable-movex-load (reg)
  (let ((regcode (get-register-code reg)) (mov-instr 8#67))
       (store-relocatable-movex                      % ld
	    (wplus2 (wshift mov-instr 26)
		    (wplus2  (wshift regcode 21) (wshift regcode 16))))))
