%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           registers.sl
% Title:          HP RISC PSL register allocation
% Author:         Winfried Neun , ZIB Berlin 
% Created:        19 September 1991
% Mode:           Lisp
% Package:        Compiler
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%	Define register names
%%
(deflist '((r0 0) (r1 1) (r2 2) (r3 3) (r4 4) (r5 5) (r6 6) (r7 7)
	   (r8 8) (r9 9) (r10 10) (r11 11) (r12 12) (r13 13) (r14 14)
	   (r15 15) (r16 16) (r17 17) (r18 18) (r19 19) (r20 20)
	   (r21 21) (r22 22) (r23 23) (r24 24) (r25 25) (r26 26)
	   (r27 27) (r28 28) (r29 29) (r30 30) (r31 31)
	   (cr11 11)
	   (sr0 0) (sr1 1) (sr2 2) (sr3 3) (sr4 4) (sr5 5) (sr 6) (sr7 7))
  '*register-code)

%%
%%	Psl friendly names of registers.
%%
(deflist
  '((argone r26) (argtwo r25) (argthree r24) (argfour r23) (argfive r22)
    (NIL r28) (st r30) (sp r30) (unix-st r30) (returnaddr r2)
    (null r0) (zero r0) 
    (bndstkptr r3) (bndstklowerbound r4) (bndstkupperbound r5)
    (ff1 r6) (ff2 r7) (ff3 r8) (ff4 r9)
    (symfncstart r12) (symvalstart r13)
    (t1 r21) (t2 r20) (t3 r19) (t4 r18) (t5 r17) (t6 r16)) '*other-name)

(put 'nil '*other-name 'r28) %doesnt work in deflist ??
(put 't   '*other-name 'r15)

(deflist
  '(    (heaplast r10) (heaptrapbound r11) (t r15)
        (bndstkptr r3) (bndstklowerbound r4) (bndstkupperbound r5)
        ) 'Registername)

(de expand-operand (asm-code)
  (cond ((atom asm-code) asm-code)
        (t   (cons (car asm-code)
               (mapcar (cdr asm-code)
                       (function (lambda(x) (to-standard-name x))))))))

(global '(*REGISTER-NAMES*))
 
(setq *REGISTER-NAMES*  '((0 r0) (1 r26) (2 r25) (3 r24) (4 r23) (5 r22)))
 
(de to-standard-name (u)
  (cond ((idp u) (cond ((get u '*other-name) (get u '*other-name))
                        (t u)))
        ((pairp u)
         (cond ((eq (car u) 'REG)
                    (cond ((numberp (cadr u))
                            (cadr (assoc (cadr u) *REGISTER-NAMES*)))
                         ((idp (cadr u)) (to-standard-name (cadr u)))
                         (t (error 0 (list "Unknown HP RISC register " u)))))
               ((member (car u) '(MKITEM WARRAY)) u)
               ((eq (car u) 'DEFERRED)   (list (to-standard-name (cadr u))))
               ((member (car u) '(INDEXED DISPLACEMENT))
                (list (car u) (to-standard-name (cadr u))
		      (to-standard-name (caddr u))
                      (to-standard-name (cadddr u))))
               ((eq (car u) '+) (cadr u))
               ((eq (car u) '-) u)
               (t (mapcar u (function to-standard-name)))))
        (t u)))
 
(de spectrum-register-p (r)
  (cond ((idp r)
         (cond ((get r 'Registername) T)
               ((get r '*other-name) (spectrum-register-p (get r '*other-name)))
               ((get r '*register-code) t)
               ((get r '*f-register-code) t)
               ((get r '*c-register-code) t)
               (t nil)))
        ((eqcar r 'REG) t)
        (t nil)))
 
(de get-register-code (reg)
      (cond ((pairp reg) (get-register-code (cadr reg)))
            ((and (get reg '*other-name)
                        (get-register-code (get reg '*other-name))))
            ((eq reg 0) 0)
            ((and (numberp reg) (lessp reg 6)) (plus2 (minus reg) 27))
            ((get reg '*register-code) (get reg '*register-code))
            (t (error 0 (list "Wrong register argument" reg))))
)
              
(de regp (r) (eqcar r 'reg))

(de fregp (r) (regp r))
