%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                       %
% File:         PC:PASS-3-CRAY                                          %
% Description:  Modifications of pass 3 for CRAY                        %
% Author:       H. Melenk (ZIB Berlin)                                  %
% Created:      12-Mar-86                                               %
% Modified:                                                             %
% Mode:         Lisp                                                    %
% Package:                                                              %
% Status:       Experimental (Do Not Distribute)                        %
%                                                                       %
%      Based on original Standard LISP Compiler by Griss and Hearn      %
%                                                                       %
% Copyright (c) 1987, Konrad Zuse Zentrum Berlin, all rights reserved   %
%                                                                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Fixed a bug in comandor (like in $pc version)

%************************************************************
%  Pass1 compile functions
%************************************************************
 
%************************************************************
%  converting (NCONS x) to (CONS x NIL) due to open coded cons
%************************************************************
(DefMacro PA1M-ncons (l) (list 'cons l nil))
 
(put 'NCONS 'pass-1-macro    'PA1M-ncons)
 
%************************************************************
%  Pass2 compile functions
%************************************************************
 
%************************************************************
%  handling of (IDAPPLY ...) and (APPLY (QUOTE atm) ...)
%************************************************************
 
(remd '&comIdApply)
 
(de &comIdApply (exp status)
 
  % COMPFN:  Look for LIST in IdApply(foo, LIST(a,,,))
  %          and produce direct call instruction
  %
 
  (let* ((fn     (second exp))
       (slow-function 'idapply)
       (args   (consed-list (third exp))))
    (if (or (eq args 'not-a-consed-list)
          (geq (length args) maxnargs&))
      (&call slow-function (rest exp) status)
      (let ((n-args (length args)))
      (setf args (append args (list fn)))
      (&comload args)
      (&attach (list '*fast-apply-load (&mkreg (plus n-args 1))))
      (&attach (list '*Id-apply  n-args))
      (setf regs& nil)
      (&remvrefs)
      ))))
 
% a new version of &regsused must handle the ID-apply cmacro
% which is introduced by this &comidapply version
 
(de &regsused (start stop regsused exitflag)
  %        scs
  % Find which registers are used.  'ALL is returned if all are used.
  % Supports INSTALLDESTORY and Reg mapping
  (if (eq start stop)
    regsused
    (prog (mac instr fn nargs)
          (setq mac (car start))
          (setq instr (car mac))
          (when (and (setq nargs
                      (cond ((eq instr '*link) (setq fn (cadr mac))
                             (cadddr mac))
                            ((or (eq instr '*linke) (eq instr '*linkef))
                             (setq fn (caddr mac)) (car (cddddr mac)))))
                     (not exitflag))
            (while (greaterp nargs 0)
              (setq regsused (adjoin (&mkreg nargs) regsused))
              (setq nargs (sub1 nargs))))
          (return (cond ((or (equal instr '*link)
                             (equal instr '*id-apply)
                          (and (equal instr '*linke) exitflag))
                         (cond ((equal fn name&) (if exitflag
                                  (&regsused (cdr start) stop regsused
                                   exitflag)
                                  'all))
                               ((and *usingdestroy
                                 (setq dest& (get fn 'destroys)))
                                (setq regsused (union dest& regsused))
                                (&regsused (cdr start) stop regsused
                                 exitflag))
                               (t 'all)))
                        ((or (equal instr '*linkf)
                          (and (equal instr '*linkef) exitflag))
                         'all)
                        (t
                         (&regsused (cdr start) stop
                          (&regsused1 regsused (cdr mac)) exitflag)))))))
 
 
(remd '&comAddressApply)
 
(de &comaddressApply (exp status)
 
  % COMPFN:  Look for LIST in addressapply(foo, LIST(a,,,))
  %          and produce direct call instruction
  %
 
  (let* ((fn     (second exp))
       (slow-function 'addressapply)
       (args   (consed-list (third exp))))
    (if (or (eq args 'not-a-consed-list)
          (geq (length args) maxnargs&))
      (&call slow-function (rest exp) status)
      (let ((n-args (length args)))
      (setf args (append args (list fn)))
      (&comload args)
      (&attach (list '*fast-apply-load (&mkreg (plus n-args 1))))
      (&attach (list '*address-apply  n-args))
      (setf regs& nil)
      (&remvrefs)
      ))))
 
(de &comapply (exp status)
 (prog (fn argss)
  (setq fn (second exp))    % test case (APPLY (QUOTE atom) ...)
  (setq argss (third exp))
  (return
    (if
      (and (pairp fn)
           (eqcar fn 'quote)
           (idp (second fn)) )
      (&comidapply (list 'idapply fn argss) status)
      (&comapply-aux exp status 'apply 'fastapply)
))) )
 
(remd 'consed-list)
 
(de consed-list (exp)
  % If exp is a list that can be decomposed by the compiler, a list
  % of its components is returned.  If not, 'NOT-A-CONSED-LIST is returned.
  (cond
   ((atom exp) 'not-a-consed-list)

   ((memq (first exp) '(list   ncons  list1  list2 list3  list4 list5
                        list6  list7  list8  list9 list10 list11
                        list12 list13 list14 list15))
          (rest exp))
 
    % the following case had to be inserted because with an open
    % coded cons ncons is no longer used in LIST expansion. Instead
    % of NCONS an explicit CONS with nil is generated. The nil
    % value is put in quote by the earlier passes
   ((and (eq (first exp) 'cons ) (equal (third exp) '(quote nil)))
       (list (second exp)))
 
 % ((eq (first exp) 'quote)  wrong handling modified; HM, 21-dec-89
 %  (second exp)
 %  )
   ((and (eqcar exp 'quote) (pairp (second exp)))
    (mapcar (second exp) (function mkquote)))
   (t 'not-a-consed-list)))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% W. Neun , ZIB Berlin changed constant for maximum gap in jumptable
%           from 5 to 10 for standard situations like case for tags
% 9.May.86  the jumpon macros produces much code, so it is better to
%           to have some jump-entries lost (a halfword)
%
(de &comcase (exp status&)
  % COMPFN: Compile CASE statement
  %         Look for special cases as JUMPON, JUMPEQ , JUMPWITHIN etc
  (prog (bottomlab regs1& jumps exps elselab high low saveregs jmps jlist
                   ranges table branchpoint& tail&)
        %  scs
        (setq bottomlab (&genlbl))
        (setq regs1& nil)
        (&comval (cadr exp) 1)
        (setq jumps (setq exps nil))
        (foreach thing on (cddr exp) do
                 (prog (lab)
                       (setq lab (&genlbl))
                       (setq jumps
                        (nconc jumps (list (list (caar thing) lab))))
                       (setq exps
                        (nconc exps (list (list lab (cadar thing)))))
                       (when (null (cdr thing))
                         (if (not (null (caar thing)))
                           (if (greaterp status& 1)
                             (progn
                               (setq regs1& (cons regs& regs1&))
                               (setq elselab bottomlab))
                             (setq exps
                              (nconc exps
                               (list (list (setq elselab (&genlbl)) ''nil)))))
                           (setq elselab lab)))))
        (setq ranges nil)
        (setq table nil)
        (foreach jmp in jumps do
                 (foreach num in (car jmp) do (if (eqcar num 'range)
                    (prog (high low)
                          (setq low (&getnum (cadr num)))
                          (setq high (&getnum (caddr num)))
                          (when (geq high low)
                            (if (lessp (difference high low) 6)
                              (for (from i low high 1)
                               (do
                                (setq table (&instbl table i (cadr jmp)))))
                              (setq ranges
                               (nconc ranges
                                (list (list low high (cadr jmp))))))))
                    (setq table (&instbl table (&getnum num) (cadr jmp))))))
        (foreach r in ranges do
                 (&attach (list '*jumpwithin (caddr r) (car r) (cadr r))))
        (while table
          (setq jmps (list (car table)))
          (setq low (setq high (caar table)))
          (setq jlist (list (cadar table)))
     %%%  (while (and (cdr table) (lessp (car (cadr table)) (plus high 5)))
          (while (and (cdr table) (lessp (car (cadr table)) (plus high 16)))
            (setq table (cdr table))
            (while (lessp high (difference (caar table) 1))
              (setq high (plus high 1))
              (setq jlist (nconc jlist (list elselab))))
            (setq high (plus high 1))
            (setq jlist (nconc jlist (list (cadar table))))
            (setq jmps (nconc jmps (list (car table)))))
          (if (lessp (length jmps) 4)
            (foreach j in jmps do
                     (&attach
                      (list '*jumpeq (cadr j) '(reg 1)
                       (list 'wconst (car j)))))
            (prog (reg)
                  (if (cdr table)
                    (progn
                      (setq reg '(reg 2))
                      %  scs
                      (&lreg reg '(reg 1)))
                    (setq reg '(reg 1)))
                  (&attach
                   (cons '*jumpon (cons reg (cons low (cons high jlist)))))
                  (setq regs& (delasc reg regs&))))
          %  scs
          (setq table (cdr table)))
        (&attjmp elselab)
        (setq saveregs regs&)
        (&setbranch)
        %      scs
        (foreach thing in exps do (progn
                   (&attlbl (car thing))
                   (setq regs& saveregs)
                   (when (cadr thing)
                     (&comval (cadr thing) status&))
                   (&nextbranch)
                   %      scs
                   (unless (&transferp codelist&)
                     (&attjmp bottomlab)
                     (setq regs1& (cons regs& regs1&)))))
        (&attlbl bottomlab)
        (setq regs& (&rmerge regs1&))
        (&unbranch)))
 
(de &comandor (exp status&)
  % COMPFN: AND and OR chains
  (prog (fn labl regsl)
        (setq fn (eq (car exp) 'and))
        (setq labl (&genlbl))
        (setq exp (cdr exp))
        (while exp
         (if (or (atom (car exp)) (not (get (caar exp) 'opentst))
                 (not (cdr exp))
                 (eq (caar exp) 'and) (eq (caar exp) 'or))
          (progn
           (&comval (car exp) (if (cdr exp)
                      1
                      status&))
           %to allow for recursion on last entry;
        %to allow for recursion on last entry;
        % BUG Busters strike again.  The following line will reset the
        % store optimization list to NIL.  This will fix the following
        % code:
        % (de foo ()
        %  (prog (x)
        %    (or (setf x 1) (setf x 2))
        %    (print x)))
        % Which was optimizing away the first setf.
           (setq slst& nil)
           (setq regsl (cons regs& regsl))
           (when (cdr exp)
             (if fn
               (&jumpnil labl)
               (&jumpt labl)))
          ) % end progn
          (let ((next-case-label (&genlbl))
                (switch&     fn)             % Used by &COMTST
                regs1&                       % Set by &COMTST
      %         branch-regs
                branchpoint&
                tail&
               )
            (&comtst (first exp) next-case-label)
            (&setbranch)
            (&comval (list 'quote (not fn)) status&)
            (&nextbranch)
            (&attjmp labl)
      %     (setq branch-regs (cons regs& branch-regs))
            (setq regs& regs1&)
            (&attlbl next-case-label)
            (&unbranch)
          ))   % let, if
           (setq exp (cdr exp)))
        (setq regs&  nil) %%%%%%%%%%%%%%%%%%%%(&rmerge regsl))
        (&attlbl labl)))
%*************************************************************
%----------------------------- Post Code Generation Fixups;
(de &pass3 ()
  % Pass 3 - optimization.
  %    The optimizations currently performed are:
  % 1. Deletion of stores not yet picked up from SLST!&.
  % 2. Removal of unreachable macros.
  % 3. A peep hole optimizer, currently only optmizing LBL macros.
  % 4. Removal of common code chains
  % 5. Changing LINK to LINKE where possible
  % 6. Squeezing out unused frame locations and mapping the stack onto
  %    the registers.
  % 7. Modify dealloc sequence for CRAY
  % Other functions of PASS3 are to tack exit code on the end and reverse
  % the code list.
  (progn 
    (foreach j in slst& do (when (car j)
               (&delmac (cdr j))))
    %    scs
    (&attlbl exitt&)
    (&attach '(*exit (*framesize)))
    (&remcode t)
    (&fixlabs)
    (&fixchains)
    (&fixlinks)
    (&remcode nil)
    (&fixfrm nil nil)
    %   scs
    (&peepholeopt)
    (&remcode nil)
    (setq codelist& (reversip codelist&))
    (setq codelist& (ModifyAllocs codelist&))
    nil))
 
%
% 1. Division of (*EXIT n) into (*DEALLOC n) and (*END n) and moving
%    *DEALLOC away from the end as far as possible in order to start
%    the frame cleanup early.
%
% 2. Division of (*DEALLOC n) into (*PREDEALLOC n) and (*POSTDEALLOC n)
%    and moving *PREDEALLOC away from *POSTDEALLOC and *END as fas as
%    possible
%
%    optimal result:
%      (*ENTRY ...)
%      (*ALLOC n)
%        . . .
%      (*        ) last reference to a frame postion
%      (*PREDEALLOC n)
%        . . .     operations with LISP registers only and *LINKs
%      (*LINK . . .) last link
%      (*POSTDEALLOC n)
%        . . .     final operations, e.g. moving NIL to (Reg 1)
%      (*END n)
%
% If *PREDEALLOC and *POSTDEALLOC are adjacent, *DEALLOC will be used
% instead
%
% Revision W. Neun, 1.September 1988
% changed modifyallocs such that dealloc will not walk above T or NIL
% to avoid conflicts in blockloading
 
(compiletime (load if-system))
(compiletime
 
   (ds IsaLink (u)
       (and (pairp u)
            (memq (car u) '(*LINK *ID-APPLY )) ))
 
)
 
 
(compiletime
 
   (ds IsaLinkE (u)
       (and (pairp u)
            (eq (car u) '*LINKE) ))
 
)
 
(compiletime
 
   (ds IsaExit (u)
       (and (pairp u)
            (eq (car u) '*EXIT) ))
 
)
 
(compiletime
 
   (ds IsaHardLink (u)
       (and (pairp u)
            (memq  (car u) '(*LINKE *CALL )) ))
)
 
(de FrameOrLabelOrAlloc (u)
 (prog (v caru)
  (return
   (cond 
      ((null u) nil)
      ((atom u) nil)
      ((eq (setq caru (car u)) 'FRAME) t)
      ((eq caru 'LABEL) t)
      ((eq caru '*ALLOC) t)
      ((and (eq caru 'quote) (or (eq (setq v (cadr u)) 'NIL)
                                 (eq v 'T))) t)
      ((FrameOrLabelOrAlloc (car u)) t)
      (t  (FrameOrLabelOrAlloc (cdr u)))   ))))
 
(fluid '(*hardalloctrace))
 
(if_system Cray-2     % not of any use here
 (de ModifyAllocs (u) u)
 
 (de ModifyAllocs (u)
     (when (not *lapopt) (return u))
     (prog(v)
                  %  (mapc u 'print)
        (setq v (ModifyAllocs1 u))
        (when v (setq u v))
        (when *hardalloctrace
             (prin2t "===============================================")
             (mapc v 'print)
             (prin2t "===============================================")
        )
        (setq v (ModifyAllocs2 u))
        (when v (setq u v))
        (when *hardalloctrace
              (mapc v 'print)
        )
        (return u)))
)
 
(de ModifyAllocs1 (u)
   (prog (v r allocsize status c1 c2 fastframe framerelevant)
      (setq fastframe t)   % assumption
      (setq framerelevant nil)
      (setq v u)
 
      (while v
             (setq r (cons (car v) r))
             (setq v (cdr v))  )
 
                           % working loop
                           % status 0: whole *DEALLOC moving
                           % status 1: *POSTDEALLOC already set
                           % status 2: *DEALLOC complete, analyzing for
                           %           fast frame
                           % status 3: fast frame analyse finished
      (setq status -1)
      (setq c1 0)
      (setq c2 -1)
      (while r
 
         (when (and (wgeq status 0) (wleq status 1))
               (setq c2 (wplus2 c2 1)))
 
         (cond
            ((eq status 2)            % moving already completed
                  (when (or (IsaLink (car r))(IsaHardLink (car r)))
                        (setq fastFrame nil)
                        (setq status 3))))
 
         (cond
            ((IsaExit (car r))
                  (when (not (or (eq status -1) (geq status 2)))
                        (setq v u)
                        (return u))
                  (setq allocsize (cdar r))
                  (setq r (cons (cons '*end allocsize) (cdr r)))
                  (setq status 0))
 
            ((IsaLinkE (car r))
                  (when (not (or (eq status -1) (geq status 2)))
                        (setq v u)
                        (return u))
                  (if (not (get (caddar r) 'OpenCode))
                      (progn
                          (setq allocsize (cdar r))
                          (setq r (cons(cons '*POSTLINKE allocsize)
                                 (cdr r)))
                          (setq allocsize (list (car allocsize)))
                          (setq status 0) )
                      (setq status 2) ) )
 
            ((IsaLink (car r))
                  (setq framerelevant t)
                  (when (and (wgeq status 0) (wleq status 1))
                        (setq c1 (wplus2 c1 1)))
                  (when (eq status 0)
                        (setq v (cons (cons '*POSTDEALLOC allocsize) v))
                        (setq status 1) ))
 
            ((FrameOrLabelOrAlloc (car r))
                  (cond
                       ((eq status 0)
                        (setq v (cons (cons '*DEALLOC allocsize) v))
                        (setq status 2) )
                       ((eq status 1)
                        (setq v (cons (cons '*PREDEALLOC allocsize) v))
                        (setq status 2) )) )    )
 
        (setq v (cons (car r) v))
        (setq r (cdr r))
       )    % while
 
%     (when (wleq c2 0) (return u))  % nothing done
 
      (when (and fastframe framerelevant
                 (wgreaterp  *framesize 0))
            (setq v (cons (cons '*UseFastFrame allocsize) v))
            (prin2t "fast frame proposed by ModifyAllocs"))
 
      (return v)
))
 
(fluid '(spec-nalloc* *fastframesize frame-exhausted-threshold*))
 
(setq *fastframesize 6)
(setq frame-exhausted-threshold* 33)
 
(de modifyAllocs2 (u)
    % tries to move *ALLOC from the top into the body
      (prog (r oldlabel newlabel inst op oldlabelcode oldlabeldone
             labelneedednew labelneededold labelmap success)
         (setq oldlabelcode u)
         (setq modifyAllocs2maxusedframe 0)
         (while (and u (not(modifyAllocs2relevant (car u))))
                (push (pop u) r))
         (when (or (null u)(not (eqcar (car u) '*ALLOC))) (return nil))
         (setq inst (pop u))
 
NEWLOOP
         (setq inst '(dummy))
         (while (and u (setq inst (pop u))
                       (and (modifyAllocs2maxusedframe inst)
                           (greaterp modifyAllocs2maxusedframe *fastframesize))
                       (not(modifyAllocs2relevant inst)))
                (push (framesubst inst) r))
         (when (greaterp modifyAllocs2maxusedframe
                         frame-exhausted-threshold*) (return NIL))
         (setq op (car inst))
         (when (and (greaterp modifyAllocs2maxusedframe *fastframesize)
                    (not (eq op '*LBL)))  % labels MUST be processed
               (if (not success) (return nil)
                   (progn
                     (push 'LATEALLOC r)
                     (push inst u)
                     (go oldloop))))
 
         (cond ((eq op '*LBL) (go newlabel))
               ((eq op '*JUMP) (go newjump))
               ((or (memq op alloc2jumps))(go newcjump))
                      % do nothin in case of opencodes
               ((and (eq op '*LINK)
                     (get (cadr inst) 'opencode)
                     (not (eq (cadr inst) 'CATCHSETUP)))
                    T)
%              ((and (eq op '*LINKE) (get (caddr inst) 'exitopencode))
 %                                          (go newjcallnever))
               ((memq op
                    '(*LINK *ID-APPLY *ADDRESS-APPLY *RETJUMP *CALL))
                              (go newlink))
               ((eq op '*DEALLOC)(go newdealloc))
               ((eq op '*PREDEALLOC)(go newpredealloc))
               ((eq op '*LINKE) (go newlinke))
               ((eq op '*JCALL-NEVER-RETURN)(go newjcallnever))
               ((eq op '*END)(go newjcallnever))
               ((eq op '*JUMPON)(return nil))
           )
         (when (not (eq op 'dummy))
               (push (framesubst inst) r))
         (when (null u) (go readytest))(go newloop)
newlabel (setq oldlabel (cadr inst))
         (setq newlabel (assoc oldlabel labelmap))
         (if newlabel
             (progn (setq newlabel (cdr newlabel))
                    (setq labelneedednew(delete oldlabel labelneedednew)))
             (progn (setq newlabel `(label ,(gensym)))
                    (push (cons oldlabel newlabel) labelmap)
                    ))
         (push `(*LBL ,newlabel) r)
         (go newloop)
newjump  (setq oldlabel (cadr inst))  % unconditional jump
         (setq newlabel (assoc oldlabel labelmap))
         (if newlabel (setq newlabel (cdr newlabel))
             (progn (setq newlabel `(label ,(gensym)))
                    (push (cons oldlabel newlabel) labelmap)
                    (push oldlabel labelneedednew)))
         (push (subst newlabel oldlabel inst) r)
              % next instruction must be an old label or EXIT
         (go postjump)
newcjump (setq oldlabel (cadr inst))  % conditional jump
         (setq newlabel (assoc oldlabel labelmap))
         (if newlabel (setq newlabel (cdr newlabel))
             (progn (setq newlabel `(label ,(gensym)))
                    (push (cons oldlabel newlabel) labelmap)
                    (push oldlabel labelneedednew)))
         (push (subst newlabel oldlabel (framesubst inst)) r)
         (go newloop)
newlink  (push 'LATEALLOC r)
         (when *hardalloctrace (prin2 "="))
         (push inst r)
         (go oldloop)
newlinke (setq success T)     % die Sache lohnt
         (push  (cons '*POSTLINKE (cdr inst)) r)
         (go postjump)
newjcallnever (push inst r)
         (go postjump)
newdealloc
         (setq success T)     % die Sache lohnt
         (when *hardalloctrace (prin2 "#"))
         (go oldloop)
newpredealloc
         (setq success T)     % die Sache lohnt (wenn auch nur maessig)
         (when *hardalloctrace (prin2 "+"))
         (push '(*ALLOC 0) r)
         (go oldloop)
 
OLDLOOP
         (setq inst '(dummy))
         (while (and u (setq inst (pop u)) (not(modifyAllocs2relevant inst)))
                (push inst r))
         (when (greaterp modifyAllocs2maxusedframe
                         frame-exhausted-threshold*) (return NIL))
         (setq op (car inst))
 
         (cond ((eq op '*LBL) (go oldlabel))
               ((eq op '*JUMP) (go oldjump))
               ((or (memq op alloc2jumps))(go oldcjump))
               ((eq op '*POSTLINKE) (go plinke))
               ((eq op '*JCALL-NEVER-RETURN)(go oldjcallnever))
               ((eq op '*END)(go oldjcallnever))
           )
         (when (not (eq op 'dummy))  (push inst r))
         (when (null u) (go readytest))(go oldloop)
oldlabel (setq oldlabel (cadr inst))
         (setq labelneededold (delete oldlabel labelneededold))
         (when (member oldlabel oldlabeldone)
               (push `(*JUMP ,oldlabel)r)
               (go readytest))
         (push oldlabel oldlabeldone)
         (push inst r)
         (go oldloop)
oldjump  (setq oldlabel (cadr inst))  % unconditional jump
         (when (not (or (member oldlabel oldlabeldone)
                        (member oldlabel labelneededold)))
               (push oldlabel labelneededold))
         (push inst r)
         (go postjump)
oldcjump (setq oldlabel (cadr inst))  % conditional jump
         (setq oldlabel (cadr inst))  % unconditional jump
         (when (not (or (member oldlabel oldlabeldone)
                        (member oldlabel labelneededold)))
               (push oldlabel labelneededold))
         (push inst r)
         (go oldloop)
oldlinke 
oldjcallnever
         (push  inst r)
         (go postjump)
 
plinke   (push  inst r)
         (go postjump)
 
postjump (when (eqcar (car u) '*LBL)
               (setq u nil)    % rest only by demand
               (go readytest))
         (go oldloop)
 
READYTEST
         (when labelneedednew
               (setq u (modifyAllocs2rest labelneedednew oldlabelcode))
               (go newloop))
         (when labelneededold
               (setq u (modifyAllocs2rest labelneededold oldlabelcode))
               (go oldloop))
         (when (not success) (return nil))
         (setq u nil)
            % reverse code and insert late allocs; now we know the
            % needed size
         (when (greaterp modifyAllocs2maxusedframe *fastframesize)
               (setq modifyAllocs2maxusedframe *fastframesize))
         (while r
             (setq inst (pop r))
             (if (not (eq inst 'LATEALLOC))
                 (push inst u)
                 (setq u
                   (modifyAllocs2lateAlloc u modifyAllocs2maxusedframe)
         )   )   )
       (setq spec-nalloc!* modifyAllocs2maxusedframe) % for lapoptloads
       (return (modifyAllocs2framereform u))
))
 
 
(de modifyAllocs2lateAlloc(r size)
    % code for the late alloc in reverse order
      (for (from i 1 size)
            (do (push `(*MOVE (FASTFRAME ,i)(FRAME ,i)) r)))
      (push '(*ALLOC (*FRAMESIZE)) r)
      r) 
 
(de modifyAllocs2relevant(inst)
   (or (atom inst)
       (memq (car inst) '(*LINK *LINKE *EXIT *END *LBL *ID-APPLY
          *ALLOC *DEALLOC *PREDEALLOC *POSTDEALLOC *POSTLINKE *END *JUMP
          *JUMPON *CALL))
       (memq (car inst) alloc2jumps)))
 
(setq alloc2jumps '(*JUMP *JUMPEQ *JUMPNOTEQ *JUMPWLEQ *JUMPWGEQ
                    *JUMPWLESSP *JUMPWGREATERP
                    *JUMPIF *JUMPINTYPE *JUMPNOTINTYPE *JUMPTYPE
                    *JUMPNOTTYPE
                    *JUMPWGREATERP *JUMPWGEQ ))
 
(de modifyAllocs2rest(labelsneeded oldlabelcode)
       (prog (r)
          (while labelsneeded
                (when (not (member `(*LBL ,(car labelsneeded)) r))
                      (setq r (member `(*LBL ,(car labelsneeded))
                               oldlabelcode)))
                (setq labelsneeded (cdr labelsneeded)))
       (return r)))
 
 
(de modifyAllocs2maxusedframe (u)
    (cond ((atom u) nil)
          ((and (eq (car u) 'FRAME) (pairp (cdr u)) (numberp (cadr u)))
                  (when (greaterp (cadr u) modifyAllocs2maxusedframe)
                        (setq modifyAllocs2maxusedframe (cadr u))))
          (T (modifyAllocs2maxusedframe (car u))
             (modifyAllocs2maxusedframe (cdr u))))
    nil) 
 
(de modifyAllocs2framereform (u)
  (mapcar u (function modifyallocs2framereform1)))
 
(de modifyAllocs2framereform1 (u)
 (prog (caru)
   (return
    (cond ((atom u) u)
          ((eq (setq caru (car u)) 'quote) u)
          ((eq caru 'reg) u)
          ((eq caru '$local) u)
          ((eq caru '$global) u)
          ((eq caru '$fluid) u)
          ((and (eq caru 'FRAME) (pairp (cdr u)) (posintp (cadr u)))
              `(reg ,(getv &frame-tregs& (cadr u))))
          ((and (eq caru 'FASTFRAME) (pairp (cdr u)) (posintp (cadr u)))
              `(reg ,(getv &fast-frame-tregs& (cadr u))))
          (t (cons (modifyAllocs2framereform1 caru)
                       (modifyAllocs2framereform1 (cdr u))))
)) ))
 
(de framesubst (inst)
      (cond ((atom inst) inst)
            ((eq (car inst) 'quote) inst)
            ((and (eqcar inst 'frame) (pairp (cdr inst))
                  (posintp (cadr inst)))
                  (cons 'fastframe (cdr inst)))
            (t (mapcar inst (function framesubst)))))
 
(de freetregs () nil)
