%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  This file describes the format of each of the instructions.         %
%  To simplify matters, each possible format is named with an ID.      %
%  That ID is assigned one template which can be used to print the     %
%  instruction, and another template which can be used to deposit the  %
%  instruction in memory.                                              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%   The following formats are named as they are on the System/370
%     Reference Summary (Yellow card):
%
%   RR  :   R1,R2                       :  OP  R1R2
%   RX  :   R1,D2(X2,B2)                :  OP  R1X2  B2--D2--
%   RS  :   R1,R3,D2(B2)                :  OP  R1R3  B2--D2--
%   SI  :   D1(B1),I2                   :  OP  ??I2  B1--D1--
%   S   :   D2(B2)                      :  ---OP---  B2--D2--
%
%   Both of these formats are called SS by The Reference Summary:
%
%   SS1 :   D1(L,B1),D2(B2)             :  OP  -L--  B1--D1-- B2--D2--
%   SS2 :   D1(L1,B1),D2(L2,B2)         :  OP  L1L2  B1--D1-- B2--D2--
%
%   These formats are described but not named by the Reference Summary:
%
%   variant of  my name  explicit format      used by
%   ----------  -------  -------------------  -------
%        RR       RR1    R1                   (SPM)
%                 RR2    I1                   (SVC)
%                 MR1    M1,R2                (BC)
%                 MR2    R2                   (BC derivatives)
%
%        RX       MX1    M1,D2(X2,B2)         (BCR)
%                 MX2    D2(X2,B2)            (BCR derivatives)
%
%        RS       RSM    R1,M3,D2(B2)         (CLM,ICM,STCM)
%                 RS2    R1,D2(B2)            (SL.., SR..)
%
%        SS       SS3    D1(L1,B2),D2(B2),I3  (SPK)
%
%        S        S0                          (IPK,PTLB) (2 byte opcode)
%
%        S        S1     D1(B1)               (LPSW,SSM)  (1 byte opcode)
%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                       %
%  The LAPInstructionFormat is a list of the form:                      %
%                                                                       %
%    (<number of opcodes> . <field format>'s)                           %
%                                                                       %
%  If the number of received opcodes differs from the number specified, %
%   an error message will be generated.                                 %
%                                                                       %
%  Otherwise the fluid variables Operand!*1, Operand!*2, etc. are bound %
%   to the received operands.                                           %
%                                                                       %
%  Each field format is then evaluated and the result deposited.        %
%                                                                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                       %
%  The primitive evaluation functions are:                              %
%                                                                       %
%  Name of function:    Used to prepare one of these instruction fields:%
%  -------------------- ----------------------------------------------- %
%  LapEncodeRegister       <4-bit register number>                      %
%  LapEncodeImmediateByte  <8-bit immediate quantity>                   %
%  LapEncodeDispBase       <4-bit base register><12-bit displacement>   %
%  LapEncodeDispIndexBase  <4-bit index reg><4-bit base reg><12-bit disp>
%  LapEncodeFourBitMask    <4-bit mask>                                 %
%  LapEncodeLength         <4-bit length> or <8-bit length>             %
%                                                                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


(deflist '((rr (2 (lapdepositbytenibblenibble opcodenumber* 
                        (lapencoderegister operand*1) 
                        (lapencoderegister operand*2)))) 

            (rr1 (1 (lapdepositbytenibblenibble opcodenumber* 0 
                        
                        (lapencoderegister operand*1)))) 

            (rr2 (1 (lapdepositbytebyte opcodenumber* 
                        (lapencodeimmediatebyte operand*1)))) 

            (mr1 (2 (lapdepositbytenibblenibble opcodenumber* 
                        (lapencodefourbitmask operand*1) 
                        (lapencoderegister operand*2)))) 

            (mr2 (1 (lapdepositbytenibblenibble opcodenumber* 
                        (lapencodefourbitmask (get opcode* 'mask)) 
                        (lapencoderegister operand*1)))) 

            (rx (2 (lapdepositbytenibblenibble opcodenumber* 
                        (lapencoderegister operand*1) 
                        (lapencoderegister (operandindexregisterof 
                                                operand*2))) 
                      (lapdeposithalfword 
                        (lapencodedispbase operand*2)))) 

            (mx1 (2 (lapdepositbytenibblenibble opcodenumber* 
                        (lapencodefourbitmask operand*1) 
                        (lapencoderegister (operandindexregisterof 
                                                operand*2))) 
                      (lapdeposithalfword 
                        (lapencodedispbase operand*2)))) 

            (mx2 (1 (lapdepositbytenibblenibble opcodenumber* 
                        (lapencodefourbitmask (get opcode* 'mask)) 
                        (lapencoderegister (operandindexregisterof 
                                                operand*1))) 
                      (lapdeposithalfword 
                        (lapencodedispbase operand*1)))) 

            (rs (3 (lapdepositbytenibblenibble opcodenumber* 
                        (lapencoderegister operand*1) 
                        (lapencoderegister operand*2)) 
                      (lapdeposithalfword 
                        (lapencodedispbase operand*3)))) 

            (rsm (3 (lapdepositbytenibblenibble opcodenumber* 
                        (lapencoderegister operand*1) 
                        (lapencodefourbitmask operand*2)) 
                      (lapdeposithalfword 
                        (lapencodedispbase operand*3)))) 

            (rs2 (2 (lapdepositbytenibblenibble opcodenumber* 
                        (lapencoderegister operand*1) 0) 
                        
                      (lapdeposithalfword 
                        (lapencodedispbase operand*2)))) 

            (si (2 (lapdepositbytebyte opcodenumber* 
                        (lapencodeimmediatebyte operand*2)) 
                      (lapdeposithalfword 
                        (lapencodedispbase operand*1)))) 

            (ss1 (2 (lapdepositbytebyte opcodenumber* 
                        (lapencodelength operand*1 256)) 
                      (lapdeposithalfword 
                        (lapencodedispbase operand*1)) 
                      (lapdeposithalfword 
                        (lapencodedispbase operand*2)))) 

            (ss2 (2 (lapdepositbytenibblenibble opcodenumber* 
                        (lapencodelength operand*1 16) 
                        (lapencodelength operand*2 16)) 
                      (lapdeposithalfword 
                        (lapencodedispbase operand*1)) 
                      (lapdeposithalfword 
                        (lapencodedispbase operand*2)))) 

            (ss3 (2 (lapdepositbytenibblenibble opcodenumber* 
                        (lapencodelength operand*1 16) 
                        (lapencodefourbitmask operand*3)) 
                      (lapdeposithalfword 
                        (lapencodedispbase operand*1)) 
                      (lapdeposithalfword 
                        (lapencodedispbase operand*2)))) 

            (s (1 (lapdeposithalfword opcodenumber*) 
                      (lapdeposithalfword 
                        (lapencodedispbase operand*1)))) 

            (s0 (0 (lapdeposithalfword opcodenumber*))) 

            (s1 (1 (lapdepositbytebyte opcodenumber* 0) 
                      (lapdeposithalfword 
                        (lapencodedispbase operand*1)))) 

            (diagnose (2 
                   (lapdepositbytebyte opcodenumber* 
                        (lapencodeimmediatebyte operand*1)) 
                      (lapdeposithalfword operand*2))) 

            (entry nil) 

            (using nil) 

            (ds (1 (lapallocatestorage operand*1))) 
            (dc (1 (lapallocateconstant operand*1))) 

            (problem (nil (lapproblem operands*)))) 
                      % nil => any number of args, Operands!* = all args

            
         'lapinstructionformat)

(put 'entry 'opcodetype 'entry)
(put 'using 'opcodetype 'using)
(put 'ds 'opcodetype 'ds)
(put 'dc 'opcodetype 'dc)
(put 'problem 'opcodetype 'problem)

(put 'ipk 'opcodetype 's0)
(put 'spka 'opcodetype 's)

(deflist '((rr 2) 
            (rr1 2) 
            (rr2 2) 
            (mr1 2) 
            (mr2 2) 
            (rx 4) 
            (mx1 4) 
            (mx2 4) 
            (rs 4) 
            (rsm 4) 
            (rs2 4) 
            (si 4) 
            (s 4) 
            (s0 4) 
            (s1 4) 
            (diagnose 4) 
            (ss1 6) 
            (ss2 6) 
            (ss3 6)) 
%           (ENTRY 0)           %discontinuing this line per Jim
        'laplength)

% of file
