% 370-OPERANDS.RED - IBM 370 specific information for LAP and ASMOUT
%
% Author:      James McDonald
%              IMSSS
%              Stanford University
% Date:        19 September 1983
% Copyright (c) 1983 Stanford University
%

(fluid '(offsetforextrareg6*))
(fluid '(codeblockoffsets* datablockoffsets*))

(ds labelp (x) (atom x))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       Field extraction from operands                     %
%                                                                          %
%   These functions are used by both ASM and LAP.                          %
%   They return an internal object in a standardized format.               %
%   Their purpose is to simplify code at higher levels.                    %
%                                                                          %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


(de operandindexregisterof (operand) 
 (prog (expansion xreg) 
   (cond ((numberp operand) (return nil))) 
   (cond ((atom operand) 
   (progn (setq expansion (expandatomicoperand operand)) 
       (cond ((atom expansion) (return nil))) 
       (setq operand expansion)))) 

   (cond ((eqcar operand 'doubleindexed) 
        (setq xreg (cadr operand)))     % (... <xreg> <breg> <disp>)
   ((and (eqcar operand 'indexed) 
        (assoc (caddr operand) codeblockoffsets*)) 
        (setq xreg '(implicitreg baser))) (t 
   (return nil))) 
   (cond ((eqcar xreg 'baseregof) 
        (return (operandbaseregisterof (cadr xreg)))) 
   ((eqcar xreg 'baseregisterof) 
        (return (operandbaseregisterof (cadr xreg)))) 
   ((eqcar xreg 'indexregisterof) 
        (return (operandindexregisterof (cadr xreg)))) 
   ((or (atom xreg) (not (memq (car xreg) '(reg implicitreg)))) 
        (stderror (bldmsg "Illegal index register in %w" operand)))) 
   (return xreg)))
 

%       ----

(de operandbaseregisterof (operand) 
 (prog (expansion type breg) 
   (cond ((numberp operand) (return nil))) 
   (cond ((atom operand) 
   (progn (setq expansion (expandatomicoperand operand)) 
       (cond ((atom expansion) (return '(implicitreg baser)))) 
       (setq operand expansion)))) 

   (setq type (car operand)) 
   (cond ((equal type 'doubleindexed) 
        (setq breg (caddr operand)))    % (... <xreg> <breg> <disp>)
   ((or (equal type 'displacement) (equal type 'indexed) )
        (setq breg (cadr operand)))     % (... <breg> <disp>)
   ((equal type 'fncell) 
        (return '(reg symfncr)))        % (... <id>)
   ((memq type '(fluid global $fluid $global)) 
        (return '(reg symvalr)))        % (... <id>)
   ((equal type 'byteloc) 
        (return (operandbaseregisterof (cadr operand)))) 	% (.. <sym> <disp> <size>)
   ((equal type 'threebyteloc)          % ---obsolete soon---
        (return (operandbaseregisterof (cadr operand)))) 	% (.. <sym> <disp> )
   ((equal type 'lengthbased) 
        (return (operandbaseregisterof (caddr operand)))) 	% (.. <length> <sym>)
   ((memq type '(wvar warray wstring)) 
        (return '(implicitreg datar)))  % (... <id>)
   ((equal type 'extrareg) 
        (return '(implicitreg datar))) (t 	% (... <int>)
   (return nil))) 
   (cond ((eqcar breg 'baseregof) 
        (return (operandbaseregisterof (cadr breg)))) 
   ((eqcar breg 'baseregisterof) 
        (return (operandbaseregisterof (cadr breg)))) 
   ((eqcar breg 'indexregisterof) 
        (return (operandindexregisterof (cadr breg)))) 
   ((or (atom breg) (not (memq (car breg) '(reg implicitreg)))) 
        (stderror (bldmsg "Illegal base register %p in %p" breg operand)))) 
   (return breg)))
 

%       ----

(off syslisp)
(de operanddisplacementof (operand) 
 (prog (expansion type disp tmp) 
   (cond ((idp operand) 
   (progn (setq expansion (expandatomicoperand operand)) 
       % if atom Expansion then nothing
       (cond ((pairp expansion) (setq operand expansion)))))) 

   (cond ((pairp operand) (setq type (car operand)))) 

   (setq disp
     (cond ((atom operand) operand ) 
           ((or (equal type 'displacement ) (equal type 'indexed) )
                (operanddisplacementof (caddr operand))) % (... <xreg> <disp>)
           ((or (equal type 'based) (equal type 'lengthbased))
                (operanddisplacementof (caddr operand))) % (... <breg> <disp>)
           ((equal type 'doubleindexed) 
                (cond ((null (cdddr operand)) 0)
		      (t (operanddisplacementof (cadddr operand))))) 
                                                % (... <xreg> <breg> <disp>)
           ((memq type '(fncell fluid global $fluid $global)) 
                (idoffsetof (cadr operand))) 	% (... <id>)
           ((equal type 'byteloc) 
                (list 'plus2 (operanddisplacementof (cadr operand)) 
                              (caddr operand)))	% (... <sym> <disp2> <size>)
           ((equal type 'wconst) (get (cadr operand) 'wconst)) 
           ((equal type 'extrareg) 
                (plus offsetforextrareg6* 
                 (times addressingunitsperitem (difference (cadr operand) 6)))) 
           ((equal type 'difference) 
                (list 'difference (operanddisplacementof (cadr operand)) 
                                   (caddr operand))) 
           ((memq type '(plus plus2)) 
                (list 'plus2 (operanddisplacementof (cadr operand)) 
                              (caddr operand))) (t 
           (stderror (bldmsg "Can't determine displacement for %w" 
                                 operand))))) 
   (cond ((and (labelp disp) (setq tmp (assoc disp codeblockoffsets*))) 
        (setq disp (cdr tmp))) (t 
   (setq disp (resolvewconstexpression disp)))) 
   (cond ((and (numberp disp) (or (lessp disp 0) (greaterp disp 4095))) 
      (stderror (bldmsg "Illegal displacement %p in %p" disp operand)))) 
   (return disp)))
 

%       ----

(de operandimmediatebyteof (operand) 
 (prog (i) 
   (setq i (resolvewconstexpression operand)) 
   (cond ((pairp i) 
      (stderror (bldmsg "Can't determine immediate value for %p" operand)))) 
   (cond ((and (numberp i) (or (lessp i 0) (greaterp i 255))) 
      (stderror (bldmsg "Illegal immediate value %p in %p" i operand)))) 
   (return i)))
 

%       ----

(de operandbytelengthof (operand limit) 
 (prog (expansion type l) 
  (cond ((atom operand) 
  (progn (setq expansion (expandatomicoperand operand)) 
      (cond ((atom expansion) 
         (stderror (bldmsg "Can't determine length of %p" operand)))) 	% >>;
      (setq operand expansion))))       %taa
  (setq type (car operand)) 
  (setq l (cond ((equal type 'lengthbased) (cadr operand)) 	% (.. <l> <breg> <disp>)
       ((equal type 'byteloc) (cadddr operand)) (t 	% (.. <sym> <disp2> <l>)
       (stderror (bldmsg "Can't determine length of %p" operand))))) 

  (cond ((and (numberp l) (or (lessp l 0) (geq l limit))) 
     (stderror (bldmsg "Illegal length field %p in %p" l operand)))) 
  (return l)))
 

%       ----

(de operandfourbitmaskof (operand) 
 (cond ((atom operand) operand)
	(t (problem))))

%       ----

(de idoffsetof (x) 
 (if *writingfaslfile (times 4 (findidnumber x))
   (times2 4 (id2int x))))

(de expandatomicoperand (operand) 
 (prog (labeloffsetpair x offset) 
  (cond ((numberp operand) (return operand))) 

  (setq labeloffsetpair (assoc operand label-alist*))
  (when labeloffsetpair
	(setq offset (difference (cdr labeloffsetpair) (cdr *using*)))
	(if (lessp offset 4096)
			 (return `(indexed ,(caar *using*) ,offset))
	     (return `(indexed ,(cadar *using*) ,(difference offset 4096)))))

  (setq labeloffsetpair (assoc operand codeblockoffsets*)) 
  (cond (labeloffsetpair 
     (return (list 'indexed '(implicitreg baser) (cdr labeloffsetpair))))) 

  (setq labeloffsetpair (assoc operand datablockoffsets*)) 
  (cond (labeloffsetpair 
     (return (list 'indexed 'pseudoheap (cdr labeloffsetpair))))) 

  (cond ((stringp operand) 
     (progn (setq x (errorset (list 'intern operand) nil nil)) 
         (cond ((pairp x) (setq operand (car x))))))) 
  (cond ((not (idp operand)) 
     (stderror (bldmsg "ExpandAtomicOperand: strange atom: %p" operand)))) 

  (setq offset (get operand 'globalliteraloffset)) 
  (cond ((null offset) (return 'unknownatomicoperand))) 
  (return (list 'indexed '(implicitreg datar) offset))))
 

% of file
