%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         sun4-nk-spec.sl
% Title:        misc functions for sparc nonkernel
% Author:       Winfried Neun
% Created:      27 February 1989
% Status:       Experimental
% Mode:         Lisp
% Package:      Kernel
%
% (c) Copyright 1989, Konrad-Zuse-Zentrum Berlin, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime
  (flag '(halfwordsequal vectorequal wordsequal intern-equal equal-hardcase
          bothzero floateqn pair-equal fast-wordsequal )
                   'internalfunction))

(remd 'eqn)

(de eqn (u v)
  % Eq or numeric equality.
      (case (tag u)
        ((posint-tag)   (or (weq u v) (and (eq 0 u) (bothzero u v))))
        ((negint-tag)   (weq u v))
        ((floatnum-tag) (floateqn u v))
        ((fixnum-tag)
                (and (fixnp v) (weq (fixval (fixinf u)) (fixval (fixinf v)))))
        ((bignum-tag)   (and (bigp v) (wordsequal u v)))
        (nil (eq u v))))

(de floateqn (u v)
      (or (and (floatp v)
               (weq (wgetv (fltinf u) 1) (wgetv (fltinf v) 1)))
          (bothzero u v)))

(de bothzero (u v) (and (zerop u) (zerop v)))

(remd 'equal)

(compiletime (setq &fillframeholes nil))

(lap '((*entry equal expr 2)
       (*alloc 0)
       (*jumpeq (label yes) (reg 1)(reg 2))
       (*move (reg 1)(reg 3))
       (*move (reg 2)(reg 4))
       (*field (reg 3)(reg 3) (wconst tagstartingbit) (wconst tagbitlength))
       (*field (reg 4)(reg 4) (wconst tagstartingbit) (wconst tagbitlength))
       (*jumpnoteq (label no) (reg 3)(reg 4))
       (*jumpeq    (label no) (reg 4)(wconst id-tag))
       (*jumpeq    (label no) (reg 4)(wconst posint-tag))
       (*jumpeq    (label no) (reg 4)(wconst negint-tag))
       (*jumpnoteq (label hardcase) (reg 4) (wconst pair-tag))
	     % both are pairs now
       (*linke 0 pair-equal expr 2)
    hardcase
       (*linke 0 equal-hardcase expr 3)
    yes
       (*move (quote T)(reg 1))
       (*exit 0)
    no
       (*move (quote nil)(reg 1))
    ex
       (*exit 0)))


(lap '((*entry intern-equal expr 2)
       (*alloc 0)
       (*jumpeq (label yes) (reg 1)(reg 2))
       (*move (reg 1)(reg 3))
       (*move (reg 2)(reg 4))
       (*field (reg 3)(reg 3) (wconst tagstartingbit) (wconst tagbitlength))
       (*field (reg 4)(reg 4) (wconst tagstartingbit) (wconst tagbitlength))
       (*jumpnoteq (label no) (reg 3)(reg 4))
       (*jumpeq    (label no) (reg 4)(wconst id-tag))
       (*jumpeq    (label no) (reg 4)(wconst posint-tag))
       (*jumpeq    (label no) (reg 4)(wconst negint-tag))
       (*jumpnoteq (label hardcase) (reg 4) (wconst pair-tag))
	     % both are pairs now
       (*linke 0 pair-equal expr 2)
    hardcase
       (*linke 0 equal-hardcase expr 3)
    yes
       (*move (quote T)(reg 1))
       (*exit 0)
    no
       (*move (quote nil)(reg 1))
    ex
       (*exit 0)))

(lap '((*entry pair-equal expr 2)
       (*alloc 2)
       (*move (reg 1)(reg 3))
       (*move (reg 2)(reg 4))
    loop
       (*field (reg 3)(reg 3)infstartingbit infbitlength)
       (*move (memory (reg 3) 0) (reg 1))       
       (*field (reg 4)(reg 4)infstartingbit infbitlength)
       (*move (memory (reg 4) 0) (reg 2))
       (*move (reg 3)(frame 1)) 
       (*move (reg 4)(frame 2))
       (*jumpeq (label car-eq)(reg 1)(reg 2))
       (*link intern-equal expr 2)
       (*jumpeq (label exit)(quote nil)(reg 1))
       (*move (frame 2)(reg 4))
       (*move (frame 1)(reg 3))
   car-eq
       (*move (memory (reg 4) addressingunitsperitem) (reg 4))
       (*move (memory (reg 3) addressingunitsperitem) (reg 3))    
       (*jumpeq (label cdr-eq)(reg 3)(reg 4))   % cdr's are eq
       (*jumpnottype (label no-pair) (reg 4) pair-tag)
       (*jumpnottype (label cdr-notequal) (reg 3) pair-tag)
       (*jump (label loop))
    no-pair
       (*move (reg 4)(reg 1))
       (*move (reg 3)(reg 2))
       (*linke 2 intern-equal expr 2)
    cdr-notequal
       (*move (quote nil)(reg 1))
       (*exit 2)
    cdr-eq
       (*move (quote t)(reg 1))
    exit
       (*exit 2)
))

(de equal-hardcase (u v tg)
  % Structural equality for atoms other than ID's
  % here we kow that the items are not eq, but tags are equal
      (case tg
	((evector-tag) (evectorequal u v))
	((vector-tag)  (vectorequal u v))
	((string-tag)  (stringequal u v))
	((bytes-tag) (stringequal u v))
	((floatnum-tag) (floateqn u v))
	((fixnum-tag)
	      (weq (fixval (fixinf u)) (fixval (fixinf v))))
	((bignum-tag) (wordsequal u v))
	((words-tag)  (wordsequal u v))
	((halfwords-tag) (halfwordsequal u v))
	(nil nil))) 

(remd 'wordsequal)

(de wordsequal (u v)
  (prog (s1 i)
        (setq u (wrdinf u))
        (setq v (wrdinf v))
        (unless (weq (setq s1 (wrdlen u)) (wrdlen v)) (return nil))
        (setq i 0)
   loop
        (when (wgreaterp i s1)    (return t))
        (unless (weq (wrditm u i) (wrditm v i))  (return nil))
        (setq i (wplus2 i 1))
        (go loop)))

(remd 'halfwordsequal)

(de halfwordsequal (u v)
  (prog (s1 i)
        (setq u (halfwordinf u))
        (setq v (halfwordinf v))
        (unless (wgreaterp (setq s1 (halfwordlen u)) (halfwordlen v))
          (return nil))
        (setq i 0)
   loop
        (when (wgreaterp i s1)        (return t))
        (unless (weq (halfworditm u i) (halfworditm v i)) (return nil))
        (setq i (wplus2 i 1))
        (go loop)))

(remd 'vectorequal)

(de vectorequal (u v)
  % Vector equality without type check
  (prog (len i)
        (setq u (vecinf u))
        (setq v (vecinf v))
        (setq len (veclen u))
        (when (wneq len (veclen v))  (return nil))
        (setq i 0)
   loop
        (when (wgreaterp i len)      (return t))
        (unless (equal (vecitm u i) (vecitm v i))       (return nil))
        (setq i (wplus2 i 1))
        (go loop)))

(remd 'evlis)

(de evlis(u)
      (prog (result pointer)
           (when (not (pairp u)) (return nil))
           (setq pointer (setq result (cons (eval (car u)) nil)))
       loop
           (setq u (cdr u))
           (when (not (pairp u)) (return result))
           (rplacd pointer (cons (eval (car u)) nil))
           (setq pointer (cdr pointer))
           (go loop)))
