%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:ENVIRONMENT.SL
% Description:  Hardware and Software Information Variables
% Author:       Martin Griss
% Created:      11-Apr-84
% Modified:     12-Sep-84 11:56:34 (Brian Beach)
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
% Compile to:   PL:ENVIRONMENT.B
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 16-Dec-85 08:30 (Brian Beach)
%  Changed COMPILE TO, and added comments.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% This file defined the macro (DEFINE-GLOBAL) which is used to define
% the global variables which describe the environment.
%
% The rest of the file contains examples of the environment variables which
% should be set up on a system--by-system basis.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 
% (Were functions in CommonLISP in case code migrates, or need to
% "compute" from present hardware, etc..)
 
% The following global variables should denote either a string or NIL.
% These variables are normally initialized upon startup
% possibly from a standard configuration file.
 
(dm define-global (vv)
  `(progn (compiletime(global '(,(cadr vv))))
        (setq ,(cadr vv) ,(caddr vv))))
 
% A short string indicating the location of the computer.
 
 (define-global *site-name* "ZIB")
 
% A longer string indicating the location of the computer.
 
 (define-global *long-site-name*
         "Konrad-Zuse-Zentrum fuer Informationstechnik Berlin"
 )
 
% A string describing the currently running HPSL software.
%/CommonLISP used only for "supporting" software, eg TOPS-20
%/Provided *lisp-implementation-type*  for "HPSL" vs "PSL" vs "?"
 
 (define-global *software-type* "COS PSL 3.4")
 
% A string describing the version of the currently running HPSL software.
%/ Common LISP reserved only for supporting software (eg. "HP Ether-Hacked")
%/ used *lisp-implementation-version* for our "3.3 10-April-1984"
 
 (define-global *software-version* "3.4 21-Jan-1986")
 
% A string describing the generic name of the hardware executing HPSL.
 
 (define-global *machine-type* "CRAY ")
 
% A string describing the model of the hardware executing HPSL.
%/ Also mention micro-code release etc.
 
 (define-global *machine-version* "Model 1 / X MP")
 
% A string describing the particular machine executing HPSL.
 
 (define-global *machine-instance* "ZIB")
 
