%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXU:MULS.SL
% Description:  Support for BIGNUM package with double INUM operations
%               Lap source for HP-RISC Processor
% Author:       H. Melenk, W. Neun
% Created:      25 January 1989
% Modified:
% Mode:         Lisp
% Package:      Utilities
% Status:       Experimental
%
% (c) Copyright 1989, 1991 Konrad-Zuse-Zentrum, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% HP RISC version by W. Neun, ZIB Berlin , December 1991
%
% These are the open codes which embed the extended arithmetic
% operations such that they are accessible from compiled LISP top level.
% They map the 2*30 bit operations needed by BIG to the 2*32 bit operations
% of the processor by shifting etc.

(fluid '(*second-value* doublearray*))

(setq doublearray* (wand (gtwarray 4) -8))

(de wxtimes2 (x y)
	(wddtimes doublearray* x y)
        (setq *second-value* (wgetv doublearray* 0))
	(wgetv doublearray* 1))

(de wtimesdouble (x y)
    (prog (aa bb)
	(wdtimes doublearray* x y)
	(setq *second-value* (wor (wshift (wgetv doublearray* 0) 2)
				  (wshift (wgetv doublearray* 1) -30)))
	(return (wshift (wshift (wgetv doublearray* 1) 2 ) -2))))

% dividing a double word - word dividend

(lap '((*entry wquotientdouble expr 3)
      % called with a double length number in params 1 and 2
      % and a single length number in par 3.
      % Result is the single length quotient.
      % the remainder is placed in a fluid variable.

      % Code is taken from the Assemble Language Ref Man, p 5.6
      % where one can find some comments too.

      % adjusting the words first

	(*alloc 0)
        (*MOVE (reg 1) (reg 4))
        (*WSHIFT (reg 1) -2)
        (*WSHIFT (reg 4) 30)
        (*WOR (reg 2) (reg 4))

	(movb (compl >= )  (reg 1) (reg 5) (label check_mag))
	(add	(reg r0) (reg 2) (reg 4))
	(sub	(reg r0) (reg 4) (reg 4))
	(subb	(reg r0) (reg 5) (reg 5))
  check_mag
	(subt (compl = ) (reg r0) (reg 3) (reg t1))

	(ds	(reg r0) (reg t1) (reg r0))
	(add	(reg 4) (reg 4) (reg 4))
	(ds	(reg 5) (reg 3) (reg 5))  % 1
	(addc	(reg 4) (reg 4) (reg 4))
	(ds	(reg 5) (reg 3) (reg 5))  % 2
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 3
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 4
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 5
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 6
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 7
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 8
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 9
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 10
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 11
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 12
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 13
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 14
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 15
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 16
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 17
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 18
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 19  
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 20
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 21
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 22
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 23
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 24
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 25
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 26
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 27
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 28
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 29
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 30
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 31
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 32
	(addc	(reg 4) (reg 4) (reg 4))
	(addbf (compl < ) (reg 5) (reg r0) (label finish))
        (nop)
	(add (compl < ) (reg 3) (reg r0) (reg r0))
	(add (compl tr) (reg 5) (reg 3) (reg 5))
	(addl	(reg 5) (reg t1) (reg 5))
    finish
	(add (compl >= ) (reg 1) (reg r0) (reg r0))
	(sub	(reg r0) (reg 5) (reg 5))
	(xor (compl >= ) (reg 1) (reg 3) (reg r0))
	(sub	(reg r0) (reg 4) (reg 4))
	(*move	(reg 4) (reg 1))
	(*move	(reg 5) ($fluid *second-value*))
	(*exit 0)))


(lap '((*entry wdivide expr 2)
        (*alloc 0)
	(*move (reg 2) (reg 3))
	(*move (reg 1) (reg 2))
	(extrs (reg 2) 0 1 (reg 1))
	(movb (compl >= )  (reg 1) (reg 5) (label check_mag))
	(add	(reg r0) (reg 2) (reg 4))
	(sub	(reg r0) (reg 4) (reg 4))
	(subb	(reg r0) (reg 5) (reg 5))
  check_mag
	(subt (compl = ) (reg r0) (reg 3) (reg t1))

	(ds	(reg r0) (reg t1) (reg r0))
	(add	(reg 4) (reg 4) (reg 4))
	(ds	(reg 5) (reg 3) (reg 5))  % 1
	(addc	(reg 4) (reg 4) (reg 4))
	(ds	(reg 5) (reg 3) (reg 5))  % 2
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 3
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 4
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 5
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 6
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 7
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 8
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 9
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 10
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 11
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 12
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 13
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 14
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 15
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 16
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 17
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 18
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 19  
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 20
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 21
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 22
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 23
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 24
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 25
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 26
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 27
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 28
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 29
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 30
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 31
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 32
	(addc	(reg 4) (reg 4) (reg 4))
	(addbf (compl < ) (reg 5) (reg r0) (label finish))
        (nop)
	(add (compl < ) (reg 3) (reg r0) (reg r0))
	(add (compl tr) (reg 5) (reg 3) (reg 5))
	(addl	(reg 5) (reg t1) (reg 5))
    finish
	(add (compl >= ) (reg 1) (reg r0) (reg r0))
	(sub	(reg r0) (reg 5) (reg 5))
	(xor (compl >= ) (reg 1) (reg 3) (reg r0))
	(sub	(reg r0) (reg 4) (reg 4))
	(*move	(reg 4) (reg 1))
	(*move	(reg 5) ($fluid *second-value*))
	(*exit 0)))

(put 'addcarry 'opencode
       '(
         (*MOVE ($FLUID carry*) (reg 2))
         (*WPLUS2 (reg 1)(reg 2))
         (*MOVE (reg 1)(reg 3)) (*WSHIFT (reg 3) -30)
         (*MOVE (reg 3) ($FLUID carry*))
         (*WSHIFT (reg 1) 2)
         (*WSHIFT (reg 1) -2)))

(put 'subcarry 'opencode
       '(
         (*MOVE ($FLUID carry*) (reg 2))
         (*WDIFFERENCE (reg 1)(reg 2))
         (*MOVE (reg 1) (reg 2)) (*WSHIFT (reg 2) -31)
         (*MOVE (reg 2) ($FLUID carry*))
         (*WSHIFT (reg 2) 30)
         (*WPLUS2 (reg 1) (reg 2))
       ))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%   functions supporting word arithmetic with bbase*
%   bbase* is too large to be stored in a fluid cell;
%   so we perform operations with bbase* by shift and mask
%   instructions.

%   get the value for bbase* into a register
(put 'bbase** 'opencode
      '((*MOVE 1 (reg 1))
        (*WSHIFT (reg 1) 30)))
%     '((LDIL  16#100000 (reg 1))))

%   get the value for logicalbits* into a register
(put 'logicalbits** 'opencode
      '((*MOVE -1 (reg 1))
        (*WSHIFT (reg 1) -2)))

%   calculate (remainder x bbase*)
(put 'remainder-bbase 'opencode
      '((*WSHIFT (reg 1) 2)
        (*WSHIFT (reg 1) -2)))

%   calculate (quotient x bbase*)
(put 'quotient-bbase 'opencode
      '((*WSHIFT (reg 1) -30)))

(put 'bbase** 'destroys         '((reg 1)))
(put 'logicalbits** 'destroys   '((reg 1)))
(put 'remainder-bbase 'destroys '((reg 1)))
(put 'quotient-bbase 'destroys  '((reg 1)))
