%
% MBX_SYSCALLS.RED -calls to the MBX (MAILBOX) IPC server. 
%
% Author:      G. Q. Maguire Jr.
%              Computer Science Dept.
%              University of Utah
% Date:        2 November 1982
% Copyright (c) 1982 G. Q. Maguire Jr.
%
% 14-MAR-84, JWP, Fixed handling of returned status.
%
% ---------------------------------------------------------------------
% MBX related routines for the Mailbox
% ---------------------------------------------------------------------

(de mbx_close (handle) 
% Closes the mailbox pointed to by the handle and
% returns the status.
(prog (status) 

  (putarg 1 handle '(integer32))        % 32 bit Pointer to mailbox
  (cleararg 2)                          % Reset space for status.

  (syscall 'mbx_$close 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))



(de mbx_cond_get_rec_chan (handle channel bufferptr bufferlen) 
% Attempts ro get a mailbox record.
(prog (returnlength message status) 

  (putarg 1 handle '(integer32))        % 32 bit Pointer to mailbox
  (putarg 2 channel '(integer16))       % The channel  number to read
  (putarg 3 bufferptr '(vector16))      % Pointer to message buffer.
  (putarg 4 bufferlen '(integer32))     % Length of the buffer 
  (cleararg 5)                          % Return poiinter;
  (cleararg 6)                          % Return length.
  (cleararg 7)                          % Reset space for status.

  (syscall 'mbx_$cond_get_rec_chan 7) 

  (setq status (getarg 7 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) 
            (setq returnlength (getarg 6 '(integer16) '(inum))) 
            % convert to LISP string
            (setq message (getarg 5 (cons 'vector16 returnlength) '(halfwords))) 
            (return message))) (t 
         
     (return (apolloerror status))))))



(de mbx_create_server (name namelen bufsize maxchan) 
% Creates and  Opens a servers mailbox, returns handle to mailbox.
(prog (handle status) 

  (putarg 1 name '(string))             % a mailbox name
  (putarg 2 namelen '(integer16))       % Length of name
  (putarg 3 bufsize '(integer16))       % Size of mailbox
  (putarg 4 maxchan '(integer16))       % Max channels to mailbox
  (cleararg 5)                          % Return poiinter  to box
  (cleararg 6)                          % Return status.


  (syscall 'mbx_$create_server 6) 

  (setq status (getarg 6 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) 
            (setq handle (getarg 5 '(integer32) '(fixn))) 
            (return handle))) (t 
         
     (return (apolloerror status))))))



(de mbx_deallocater (handle channel) 
% Releases a channel.
(prog (status) 

  (putarg 1 handle '(integer32))        % a mailbox handle
  (putarg 2 channel '(integer16))       % mailbox channel number
  (cleararg 3)                          % Return status.


  (syscall 'mbx_$deallocator 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))



(de mbx_get_conditional (handle bufptr buflen) 
% Gets a mailbox record if there is one.
(prog (retptr retlen status) 

  (putarg 1 handle '(integer32))        % Pointer to mailbox
  (putarg 2 bufptr '(integer32))        % Pointer to message buff
  (putarg 3 buflen '(integer32))        % Size of message buffer.
  (cleararg 4)                          % Return poiinter  to data
  (cleararg 5)                          % Return langth of data.
  (cleararg 6)                          % Return status.


  (syscall 'mbx_$get_conditional 6) 

  (setq status (getarg 6 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) 
            (setq retlen (getarg 5 '(integer32) '(fixn))) 
            (setq retptr (getarg 4 (cons 'vector16 retlen) '(halfwordvector))) 
            (return retptr))) (t 
         
    (return (apolloerror status))))))



(de mbx_get_ec (handle key) 
% Gets the eventcount address of a mailbox.
(prog (ecpointer status) 

  (putarg 1 handle '(integer32))        % Pointer to mailbox.
  (putarg 2 key '(integer16))           % Mailbox key.
  (cleararg 3)                          % Pointer to EC array.??
  (cleararg 4)                          % Return status.


  (syscall 'mbx_$get_ec 4) 

  (setq status (getarg 4 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) 
            (setq ecpointer (getarg 3 '(integer32) '(integer32))) 
            (return ecpointer))) (t 
         
     (return (apolloerror status))))))



(de mbx_get_rec (handle bufptr buflen) 
% Gets a record from the mailbox.
(prog (retptr retlen status) 

  (putarg 1 handle '(integer32))        % Pointer to mailbox
  (putarg 2 bufptr '(integer32))        % Pointer to message buff
  (putarg 3 buflen '(integer32))        % Size of message buffer.
  (cleararg 4)                          % Return poiinter  to data
  (cleararg 5)                          % Return langth of data.
  (cleararg 6)                          % Return status.


  (syscall 'mbx_$get_rec 6) 

  (setq status (getarg 6 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) 
            (setq retlen (getarg 5 '(integer32) '(fixn))) 
            (setq retptr (getarg 4 '(integer32) '(fixn))) 
            (return retptr))) (t 
         
     (return (apolloerror status))))))


(de mbx_get_rec_chan (handle chan bufptr buflen) 
% Gets a mailbox record.
(prog (retptr retlen status) 

  (putarg 1 handle '(integer32))        % Pointer to mailbox
  (putarg 2 chan '(integer16))          % Channel number.
  (putarg 3 bufptr '(integer32))        % Pointer to message buff
  (putarg 4 buflen '(integer32))        % Size of message buffer.
  (cleararg 5)                          % Return poiinter  to data
  (cleararg 6)                          % Return langth of data.
  (cleararg 7)                          % Return status.


  (syscall 'mbx_$get_rec_chan 6) 

  (setq status (getarg 7 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) 
            (setq retlen (getarg 6 '(integer32) '(fixn))) 
            (setq retptr (getarg 5 (cons 'vector16 retlen) '(halfwords))) 
            (return retptr))) (t 
         
     (return (apolloerror status))))))



(de mbx_open (name namelen bufptr buflen) 
% Opens a client channel to a mailbox.
(prog (handle status) 

  (putarg 1 name '(string))             % a mailbox name
  (putarg 2 namelen '(integer16))       % Length of name
  (putarg 3 bufptr '(integer32))        % Data buffer pointer
  (putarg 4 buflen '(integer32))        % Amount of data
  (cleararg 5)                          % Return poiinter  to box
  (cleararg 6)                          % Return status.


  (syscall 'mbx_$open 6) 

  (setq status (getarg 6 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) 
            (setq handle (getarg 5 '(integer32) '(fixn))) 
            (return handle))) (t 
         
     (return (apolloerror status))))))



(de mbx_put_chr (handle bufptr buflen) 
% Sends a partial record from a client.
(prog (status) 

  (putarg 1 handle '(integer32))        % Pointer to mailbox
  (putarg 2 bufptr '(integer32))        % Pointer to message buff
  (putarg 3 buflen '(integer32))        % Size of message buffer.
  (cleararg 4)                          % Return status.


  (syscall 'mbx_$put_chr 4) 

  (setq status (getarg 4 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) (return status))) (t 
     (return (apolloerror status))))))



(de mbx_put_rec (handle bufptr buflen) 
% Puts a record in  the mailbox.
(prog (status) 

  (putarg 1 handle '(integer32))        % Pointer to mailbox
  (putarg 2 bufptr '(integer32))        % Pointer to message buff
  (putarg 3 buflen '(integer32))        % Size of message buffer.
  (cleararg 4)                          % Return status.


  (syscall 'mbx_$put_rec 4) 

  (setq status (getarg 4 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) (return status))) (t 
     (return (apolloerror status))))))



(de mbx_put_rec_cond (handle bufptr buflen) 
% Puts a record in  the mailbox if there is room.
(prog (status) 

  (putarg 1 handle '(integer32))        % Pointer to mailbox
  (putarg 2 bufptr '(integer32))        % Pointer to message buff
  (putarg 3 buflen '(integer32))        % Size of message buffer.
  (cleararg 4)                          % Return status.


  (syscall 'mbx_$put_rec_cond 4) 

  (setq status (getarg 4 '(integer32) '(fixn))) 	% 32 bit status code

  (cond ((equal status 0) 
    (progn (apolloresetstatus) (return status))) (t 
     (return (apolloerror status))))))

