
%               This is the SYSCALL Package.
%        The proposal is that there be N+1 XSysCall routines
% [XSysCall0, ... XSysCallN] which take the indicated number of parameters.
%
% Authors:   Steve Lowder and G.Q. Maguire
%            Computer Science dept.
%            University of Utah
%  Date:     October 1982
%
% NOTE: The workings of this and of the syscall packages using this depend
% HEAVILY on their being GTWRDS, PutHalfWord, GetHalfWord, PutWord, GetWord, etc.
% functions available for use by an interpreted system.  These are not standard
% psl but are defined specificly in System-Faslout.red for the apollo.
%
% Modified 05-Apr-87 by Hucka.  Un-rlisp'ed and converted to 3.4.
% Modified 23-Aug-86 by JWP, added support for calling C (putValue)
% Modified 20-Jun-84 by JW, included *ForceAegisLookup flag to ease 
%  debugging of dynamicly loaded libraries.
% Modified 14-Mar-84 by JW to handle returned status codes better.
%
% Modified 26-Aug-83 by JW to use generic car; this avoids tag stripping
% problems on DN300 nodes.
% Modified 6-Dec-82 by J.W. Peterson to handle capabilities for complex Pascal
% data types.
%
% Modifed 10-Jan-82 by JW for Aegis 5.0 (changed address of KG_$LOOKUP in
%   AegisEntry)
%            
% Modified 8-mar-83 by JW for Steve's latest rebuild (AegisEntry now in kernel,
%          outbound floats repaired).

(compiletime (load syslisp))            % just to make sure things like INUMP
                                        % work.  (wierd).

(fluid '(callargumentblock 
        calladdressblock 
        lastapolloerror* 
        apolloprinterror* 
        *forceaegislookup))


(loadtime (progn 
  (setq callargumentblock (gtwarray 16))  % Used for argument passing.
  (setq calladdressblock (gtwarray 16)) 
  (setq lastapolloerror* 0)               % Last error status found.
  (setq apolloprinterror* t)              % If not NIL then print errors.
  (setq *forceaegislookup nil) nil))      % If T then look up the entry for
                                          % each  call (useful for debugging
				          % dynamically loaded stuff).         

% AddressOfAegisRoutine := AegisEntry(EntryPointName)
%   LookUpAegisEntry(EntryPointName) may be called to
%   get the address and remember it for future use, on the property list of the
%   EntryPointName under the indicator 'AegisEntryAddress.

(de lookupaegisentry (entrypointname) 
  (prog (addressofaegisroutine) 
   (cond ((idp entrypointname) 
          (cond ((and (not *forceaegislookup) 
		      (setq addressofaegisroutine
			    (get entrypointname 'aegisentryaddress))) 
		 (return addressofaegisroutine)) 
		((neq (setq addressofaegisroutine
			    (aegisentry (concat (id2string entrypointname) "                              "))) 0) 
		 (progn
		   (put entrypointname 'aegisentryaddress 
			addressofaegisroutine) 
		   (return addressofaegisroutine)))
		(t (stderror (bldmsg "No Aegis entry point for %p%n" 
				     entrypointname)))))
	 (t (stderror
	     (bldmsg "LookUpAegisEntry requires the entry name as an ID %s%n" 
		     entrypointname))))))


%   *** NOW BUILT INTO KERNEL ***
%
%lap '(
%(!*Entry AegisEntry expr 1)
% (!*WPLUS2 (reg 1) 4)
% (!*PUSH (reg 1))
% (!*MOVE (immediate 16#00C52C) (reg a0))  % updated address of KG_$LOOKUP for
% (jsr (displacement (reg a0) 0))          % Aegis 5.0
% (move!.l (fluid NIL) (REG NIL))
% (!*MOVE 0 (REG a6))
% (!*WPLUS2 (reg st) 4)
% (!*MOVE (reg a0)(reg 1))
% (rts)
% );
%
%
(on syslisp)


% SysCallLoadArgBlockLoad puts into the appropriate SysCallArgumentBlock
% slot the value of the argument. It must make sure the pointer of the
% addressblock is pointing to it.

(de syscallloadargblock32 (argumentnumber argument syscallargumentblock) 
  (prog nil 
        (cleararg argumentnumber) 
        (sys2int (setf (wgetv syscallargumentblock argumentnumber)
		       (int2sys argument)))))
  
(de syscallloadargblock16 (argumentnumber argument syscallargumentblock) 
  (prog nil 
        (cleararg argumentnumber) 
        (sys2int (setf (wgetv syscallargumentblock argumentnumber)
		       (wshift (int2sys argument) 16)))))
  
(de syscallunloadargblock32 (argumentnumber syscallargumentblock) 
  (sys2int (wgetv syscallargumentblock argumentnumber)))

(de syscallunloadargblock16 (argumentnumber syscallargumentblock) 
  (sys2int (wshift (wgetv syscallargumentblock argumentnumber) -16)))

(de syscallpushargaddress (argumentnumber syscallargumentblock syscalladdressblock) 
    (setf (wgetv syscalladdressblock argumentnumber)
	  (wplus2 syscallargumentblock (wtimes2 argumentnumber 4))))

(de syscallloadaddressblock (argumentnumber argument syscalladdressblock) 
   (sys2int (setf (wgetv syscalladdressblock argumentnumber)
		  (getinf argument))))

% This exists for calling C routines, since it's call by referance instead
% of call by value.  Hence, no tag stripping.
(de syscallloadvalue (argumentnumber argument syscalladdressblock) 
   (sys2int (setf (wgetv syscalladdressblock argumentnumber) argument)))

(off syslisp)


%lap '(         % Older version; did not strip tag.
%(!*Entry ClearArg expr 1)
%         (!*Move (fluid CallArgumentBlock) (reg a0))
%         (!*MOVE (reg 1)(reg d1))
%         (lsl!.l (immediate 2)(reg d1))
%         (!*Wplus2 (reg a0)(reg d1))
%         (!*Move 0 (displacement (reg a0) 0))
%         (!*Move (fluid CallAddressBlock) (reg 1))
%         (!*Move (reg a0) (indexed (reg d1) (displacement (reg 1) 0)))
%         (!*Exit 0)
%);

% My recollection isn't 100%, but I *think* these are to keep Lap informed
% when the tag stripping is done; the values have to be tucked into the
% places lap looks for them.

(compiletime
  (progn 
    (de makeaconstonmonday (cn val) 
      (progn (put cn 'wconst val) 
	     (set cn val))) 

(makeaconstonmonday 'infstartingbit 8) 
(makeaconstonmonday 'infbitlength 24) nil))

(lap '((*entry cleararg expr 1) 
       (*move (fluid callargumentblock) (reg a0)) 
       (*move (reg 1) (reg d4)) 
       (lsl!.l (immediate 2) (reg d4)) 
       (*wplus2 (reg a0) (reg d4)) 
       (*move 0 (displacement (reg a0) 0)) 
       (*move (fluid calladdressblock) (reg 1)) 
       (*field (reg d4) (reg d4) infstartingbit infbitlength) 
       (*move (reg 1) (reg a1))
       (*move (reg a0) (indexed (reg d4) (displacement (reg a1) 0))) 
       (*exit 0)))


%lap '(
%(!*Entry SysCallUnLoadAddressBlock expr 1)
%         (!*Move (reg 1)(reg d1))                 % Older version, did
%         (lsl!.l (immediate 2) (reg d1))          % not strip tag.
%         (!*Move (fluid CallAddressBlock) (reg 1))
%         (!*Move (indexed (reg d1) (displacement (reg 1) 0)) (reg 1))
%         (!*Exit 0)
%);


(lap '((*entry syscallunloadaddressblock expr 1) 
       (*move (reg 1) (reg d1)) 
       (lsl!.l (immediate 2) (reg d1)) 
       (*move (fluid calladdressblock) (reg 1)) 
       (*field (reg d1) (reg d1) infstartingbit infbitlength) 
       (*move (reg 1) (reg a1))
       (*move (indexed (reg d1) (displacement (reg a1) 0)) (reg 1)) 
       (*exit 0)))

(lap '((*entry cleanupaftersyscall expr 0) 
%      (*move (reg d0) (reg 1)) 
       (*move (fluid nil) (reg nil))   % fix up the LISP regs
%      (*move 0 (reg a6)) 
       (*jump (entry sys2int))))


(lap '((*entry xsyscall0 expr 1) 
       (*move (reg 1) (reg a0)) 
       (*move (fluid saveda6) (reg a6)) % restore a6 before going out
       (jsr (displacement (reg a0) 0)) % jump to Aegis routine
       (*jump (entry cleanupaftersyscall))))


(lap '((*entry xsyscall1 expr 1) 
       (*push (memory (fluid calladdressblock) 4)) 
       (*move (reg 1) (reg a0)) 
       (*move (fluid saveda6) (reg a6)) % restore a6 before going out
       (jsr (displacement (reg a0) 0)) % jump to Aegis routine
       (*dealloc 1) 
       (*jump (entry cleanupaftersyscall))))


(lap '((*entry xsyscall2 expr 1) 
       (*push (memory (fluid calladdressblock) 8)) 
       (*push (memory (fluid calladdressblock) 4)) 
       (*move (reg 1) (reg a0)) 
       (*move (fluid saveda6) (reg a6)) % restore a6 before going out
       (jsr (displacement (reg a0) 0)) % jump to Aegis routine
       (*dealloc 2) 
       (*jump (entry cleanupaftersyscall))))

(lap '((*entry xsyscall3 expr 1) 
       (*push (memory (fluid calladdressblock) 12)) 
       (*push (memory (fluid calladdressblock) 8)) 
       (*push (memory (fluid calladdressblock) 4)) 
       (*move (reg 1) (reg a0)) 
       (*move (fluid saveda6) (reg a6)) % restore a6 before going out
       (jsr (displacement (reg a0) 0)) % jump to Aegis routine
       (*dealloc 3) 
       (*jump (entry cleanupaftersyscall))))

(lap '((*entry xsyscall4 expr 1) 
       (*push (memory (fluid calladdressblock) 16)) 
       (*push (memory (fluid calladdressblock) 12)) 
       (*push (memory (fluid calladdressblock) 8)) 
       (*push (memory (fluid calladdressblock) 4)) 
       (*move (reg 1) (reg a0)) 
       (*move (fluid saveda6) (reg a6)) % restore a6 before going out
       (jsr (displacement (reg a0) 0)) % jump to Aegis routine
       (*dealloc 4) 
       (*jump (entry cleanupaftersyscall))))

(lap '((*entry xsyscall5 expr 1) 
       (*push (memory (fluid calladdressblock) 20)) 
       (*push (memory (fluid calladdressblock) 16)) 
       (*push (memory (fluid calladdressblock) 12)) 
       (*push (memory (fluid calladdressblock) 8)) 
       (*push (memory (fluid calladdressblock) 4)) 
       (*move (fluid saveda6) (reg a6)) % restore a6 before going out
       (*move (reg 1) (reg a0)) 
       (jsr (displacement (reg a0) 0)) % jump to Aegis routine
       (*dealloc 5) 
       (*jump (entry cleanupaftersyscall))))

(lap '((*entry xsyscall6 expr 1) 
       (*push (memory (fluid calladdressblock) 24)) 
       (*push (memory (fluid calladdressblock) 20)) 
       (*push (memory (fluid calladdressblock) 16)) 
       (*push (memory (fluid calladdressblock) 12)) 
       (*push (memory (fluid calladdressblock) 8)) 
       (*push (memory (fluid calladdressblock) 4)) 
       (*move (reg 1) (reg a0)) 
       (*move (fluid saveda6) (reg a6)) % restore a6 before going out
       (jsr (displacement (reg a0) 0)) % jump to Aegis routine
       (*dealloc 6) 
       (*jump (entry cleanupaftersyscall))))

(lap '((*entry xsyscall7 expr 1)          %. XSysCall1(AddressOfAegisRoutine);
       (*push (memory (fluid calladdressblock) 28)) 
       (*push (memory (fluid calladdressblock) 24)) 
       (*push (memory (fluid calladdressblock) 20)) 
       (*push (memory (fluid calladdressblock) 16)) 
       (*push (memory (fluid calladdressblock) 12)) 
       (*push (memory (fluid calladdressblock) 8)) 
       (*push (memory (fluid calladdressblock) 4)) 
       (*move (reg 1) (reg a0)) 
       (*move (fluid saveda6) (reg a6)) % restore a6 before going out
       (jsr (displacement (reg a0) 0)) % jump to Aegis routine
       (*dealloc 7) 
       (*jump (entry cleanupaftersyscall))))

(lap '((*entry xsyscall8 expr 1) 
       (*push (memory (fluid calladdressblock) 32)) 
       (*push (memory (fluid calladdressblock) 28)) 
       (*push (memory (fluid calladdressblock) 24)) 
       (*push (memory (fluid calladdressblock) 20)) 
       (*push (memory (fluid calladdressblock) 16)) 
       (*push (memory (fluid calladdressblock) 12)) 
       (*push (memory (fluid calladdressblock) 8)) 
       (*push (memory (fluid calladdressblock) 4)) 
       (*move (reg 1) (reg a0)) 
       (*move (fluid saveda6) (reg a6)) % restore a6 before going out
       (jsr (displacement (reg a0) 0)) % jump to Aegis routine
       (*dealloc 8) 
       (*jump (entry cleanupaftersyscall))))

(lap '((*entry xsyscall9 expr 9) 
       (*push (memory (fluid calladdressblock) 36)) 
       (*push (memory (fluid calladdressblock) 32)) 
       (*push (memory (fluid calladdressblock) 28)) 
       (*push (memory (fluid calladdressblock) 24)) 
       (*push (memory (fluid calladdressblock) 20)) 
       (*push (memory (fluid calladdressblock) 16)) 
       (*push (memory (fluid calladdressblock) 12)) 
       (*push (memory (fluid calladdressblock) 8)) 
       (*push (memory (fluid calladdressblock) 4)) 
       (*move (reg 1) (reg a0)) 
       (*move (fluid saveda6) (reg a6)) % restore a6 before going out
       (jsr (displacement (reg a0) 0)) % jump to Aegis routine
       (*dealloc 9) 
       (*jump (entry cleanupaftersyscall))))

(lap '((*entry xsyscall10 expr 1) 
       (*push (memory (fluid calladdressblock) 40)) 
       (*push (memory (fluid calladdressblock) 36)) 
       (*push (memory (fluid calladdressblock) 32)) 
       (*push (memory (fluid calladdressblock) 28)) 
       (*push (memory (fluid calladdressblock) 24)) 
       (*push (memory (fluid calladdressblock) 20)) 
       (*push (memory (fluid calladdressblock) 16)) 
       (*push (memory (fluid calladdressblock) 12)) 
       (*push (memory (fluid calladdressblock) 8)) 
       (*push (memory (fluid calladdressblock) 4)) 
       (*move (reg 1) (reg a0)) 
       (*move (fluid saveda6) (reg a6)) % restore a6 before going out
       (jsr (displacement (reg a0) 0)) % jump to Aegis routine
       (*dealloc 10) 
       (*jump (entry cleanupaftersyscall))))

(de syscall (syscallname numberofargs) 
   (cond ((equal numberofargs 0)
	  (xsyscall0 (lookupaegisentry syscallname))) 
	 ((equal numberofargs 1)
	  (xsyscall1 (lookupaegisentry syscallname))) 
	 ((equal numberofargs 2)
	  (xsyscall2 (lookupaegisentry syscallname))) 
	 ((equal numberofargs 3)
	  (xsyscall3 (lookupaegisentry syscallname))) 
	 ((equal numberofargs 4)
	  (xsyscall4 (lookupaegisentry syscallname))) 
	 ((equal numberofargs 5)
	  (xsyscall5 (lookupaegisentry syscallname))) 
	 ((equal numberofargs 6)
	  (xsyscall6 (lookupaegisentry syscallname))) 
	 ((equal numberofargs 7)
	  (xsyscall7 (lookupaegisentry syscallname))) 
	 ((equal numberofargs 8)
	  (xsyscall8 (lookupaegisentry syscallname))) 
	 ((equal numberofargs 9)
	  (xsyscall9 (lookupaegisentry syscallname))) 
	 ((equal numberofargs 10)
	  (xsyscall10 (lookupaegisentry syscallname)))
	 (t (errorprintf "Too Many Arguments in SysCall"))))

(de initsyscall nil 
    (for (from i 0 15 1) (do (cleararg i))))

(on syslisp)


% PutArg is used to load up the argument block for SysCall arguments.

(de putarg (argumentnumber argument pascaldatatype) 
  (prog (datatype datacount floattemp sizetemp) 
	(setq datatype (car pascaldatatype)) 
	(cond ((cdr pascaldatatype)
	       (setq datacount (cdr pascaldatatype)))) 
	(cond ((inump argument) 
	       (cond ((memq datatype '(integer32 integer16)) 
		      (progn
			(cond ((weq datatype 'integer32) 
			       (syscallloadargblock32 argumentnumber argument callargumentblock))
			      (t (syscallloadargblock16 argumentnumber argument callargumentblock)))))
		     (t (errorprintf "PascalDataType not compatible with Inum in PutARG")))) 
	      ((fixp argument) 
	       (cond ((memq datatype '(integer32 integer16)) 
		      (progn
			(cond ((weq datatype 'integer32) 
			       (syscallloadargblock32 argumentnumber argument callargumentblock))
			      (t (syscallloadargblock16 argumentnumber argument callargumentblock)))))
		     (t (errorprintf "PascalDataType not compatible with Fixnum in PutARG")))) 
	      ((stringp argument) 
	       (cond ((memq datatype '(string vector8)) 
		      (syscallloadaddressblock argumentnumber (strbase argument) calladdressblock))
		     (t (errorprintf "PascalDataType not compatible with String in PutARG")))) 
	      ((bytesp argument) 
	       (cond ((memq datatype '(vector8)) 
		      (syscallloadaddressblock argumentnumber (strbase argument) calladdressblock))
		     (t (errorprintf "PascalDataType not compatible with ByteVector in PutARG")))) 
	      ((halfwordsp argument) 
	       (cond ((memq datatype '(vector32 vector16)) 
		      (syscallloadaddressblock argumentnumber (strbase argument) calladdressblock))
		     (t (errorprintf "PascalDataType not compatible with HalfWordVector  in PutARG")))) 
	      ((wrdsp argument) 
	       (cond ((memq datatype '(vector32)) 
		      (syscallloadaddressblock argumentnumber (strbase argument) calladdressblock))
		     (t (errorprintf "PascalDataType not compatible with WordVectorP in PutARGd")))) 
	      ((vectorp argument) 
	       (cond ((memq datatype '(vector64)) 
		      (progn 
			(setq sizetemp (size argument)) 
			(setq floattemp (gtwrds (wtimes2 2 sizetemp)))
			(for (from i 0 sizetemp 1)
			      (do (progn 
				    (putword floattemp (wtimes2 i 2) (getword (floatbase (fltinf (wgetv argument (wplus2 i 1)))) 0)) 
				    (putword floattemp (wplus2 (wtimes2 i 2) 1) (getword (floatbase (fltinf (wgetv argument (wplus2 i 1)))) 1))))) 
                   
			(syscallloadaddressblock argumentnumber floattemp calladdressblock)))
		     (t (errorprintf "PascalDataType not compatible with VectorP in PutARG")))) 
	      ((floatp argument) 
	       (cond ((memq datatype '(real64)) 
		      (syscallloadaddressblock argumentnumber (floatbase (fltinf argument)) calladdressblock)) 
		     ((memq datatype '(real32)) 
		      (progn (setq floattemp (gtbps 1)) 
			     (*float64to32 floattemp (floatbase (fltinf argument))) 
			     (syscallloadaddressblock argumentnumber floattemp calladdressblock)))
		     (t (errorprintf "PascalDataType not compatible with WordVectorP in PutARGd"))))
	      (t (errorprintf " Unknown Lisp Data Type in PutArg  ")))))

% PutObjPointer places a pointer to a user-reserved Pascal data type into
% the call address block.  jwp
%
(de putobjpointer (argumentnumber argument) 
  (syscallloadaddressblock argumentnumber argument calladdressblock))

%
% This load a value directly into the call address block.  It's used
% for calling routines written in C, since they are call by value
% (Hence the stuff in the call address block is used directly).  The
% tag of argument is not stripped along the way.
%
(de putvalue (argumentnumber argument) 
  (syscallloadvalue argumentnumber (int2sys argument) calladdressblock))

(de getarg (argumentnumber pascaldatatype lispobjecttype) 
  (prog (datatype objecttype floattemp 
               datacount objectcount 
               length1 namestring stringptr 
               vecptr vector final) 

    (setq datatype (car pascaldatatype)) 
    (cond ((cdr pascaldatatype)
	   (setq datacount (cdr pascaldatatype)))) 

    (setq objecttype (car lispobjecttype)) 
    (cond ((cdr lispobjecttype)
	   (setq objectcount (cdr lispobjecttype)))) 

    (cond ((weq objecttype 'inum) 
	   (cond ((memq datatype '(integer32 integer16)) 
		  (progn (return 
			  (cond ((weq datatype 'integer32) 
				 (syscallunloadargblock32 argumentnumber callargumentblock))
				(t (syscallunloadargblock16 argumentnumber callargumentblock))))))
		 (t (errorprintf "PascalDataType not compatible with Inum in GetARG"))))
	  ((weq objecttype 'fixn) 
	   (cond ((memq datatype '(integer32 integer16)) 
		  (progn (return 
			  (cond ((weq datatype 'integer32) 
				 (syscallunloadargblock32 argumentnumber callargumentblock))
				(t (syscallunloadargblock16 argumentnumber callargumentblock))))))
		 (t (errorprintf "PascalDataType not compatible with Fixnum in GetARG"))))
	  ((weq objecttype 'string) 
	   (cond ((memq datatype '(string vector8)) 
		  (progn (setq namestring (gtstr (wdifference datacount 1))) 
			 (setq stringptr (syscallunloadaddressblock argumentnumber)) 
             (setq final (wdifference datacount 1))
             (for 
               (from i 0 final 1)
		       (do 
		        (putstrbyt namestring i (getbyte stringptr i)))) 
			 (return (mkstr namestring))))
		 (t (errorprintf "PascalDataType not compatible with String in GetARG"))))
	  ((weq objecttype 'bytevector) 
	   (cond ((memq datatype '(vector8)) 
		  (return (mkbytes (syscallunloadargblock32 argumentnumber callargumentblock))))
		 (t (errorprintf "PascalDataType not compatible with ByteVector in GetARG"))))
	  ((weq objecttype 'halfwordvector) 
	   (cond ((memq datatype '(vector32 vector16)) 
		  (progn (setq vector (gthalfwords (wdifference datacount 1))) 
			 (setq vecptr (syscallunloadaddressblock argumentnumber))
             (Setq final (wdifference datacount 1))
			 (for (from i 0 final 1)
			       (do (puthalfword vector i (gethalfword vecptr i)))) 
			 (return vector)))
		 (t (errorprintf "PascalDataType not compatible with HalfWordVector  in GetARG"))))
	  ((weq objecttype 'wordvector) 
	   (cond ((memq datatype '(vector32)) 
		  (mkwrds (syscallunloadargblock32 argumentnumber callargumentblock)))
		 (t (errorprintf "PascalDataType not compatible with WordVector in GetARG"))))
	  ((weq objecttype 'lispfloat) 
	   (cond ((memq datatype '(real64)) 
		  (return (mkfltn (syscallunloadargblock32 argumentnumber callargumentblock)))) 
		 ((memq datatype '(real32)) 
                  (progn 
		    (setq floattemp (gtfltn 1)) 
		    (*float32to64 (floatbase (fltinf floattemp)) (syscallunloadargblock32 argumentnumber callargumentblock)) 
		    (return (mkfltn floattemp))))
		 (t (errorprintf "PascalDataType not compatible with LispFloat in GetARG"))))
	  (t (errorprintf " Unknown Lisp Data Type in GetArg")))))


(de getinf (x) (inf x))

(off syslisp)



% ****************** Status code handling *********************
%
% There is a global variable called LastApolloError!*
% containing the status code of the last syscall.  The
% function ApolloError sets this global, and
% if the global ApolloPrintError!* is not NIL, prints
% the respective error message.  If ApolloPrintError!* is
% NIL, then ApolloError sets LastApolloError!* silently.
%
% Functions may still return bogus data, so the
% status should be checked before using returned syscall
% values.
%
% The function ApolloResetStatus() will reset 
% LastApolloError!* if the function completed successfully.  To date,
% all have been modified to do this (whew!).  PLEASE CONTINUE THAT
% TRADITION.
%

(de stringlength (x) 
  (plus (size x) 1))

% print on the Apollo error channel the error message
% Currently the generic error handler for system calls
(de apolloerror (status) 
  (prog (kloo) 
	(setq lastapolloerror* status) 
	(cond (apolloprinterror* 
	       (progn 
		 (putarg 1 status '(integer32)) 
		 (syscall 'error_$print 1) nil))) 
	% Will probably be replaced with ERROR_$GET_TEXT eventually
	(return status)))

%
% Reset the status if everything's OK.
%
(de apolloresetstatus nil 
   (setq lastapolloerror* 0))
